# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

from volatility3.framework import objects, interfaces
from volatility3.framework import exceptions
from volatility3.framework.symbols import wrappers
from volatility3.framework import renderers
from typing import Union


class SERVICE_RECORD(objects.StructType):
    """A service record structure."""

    def is_valid(self) -> bool:
        """Determine if the structure is valid."""
        try:
            if self.Order < 0 or self.Order > 0xFFFF:
                return False

            try:
                _ = self.State.description
                _ = self.Start.description
            except ValueError:
                return False
        except exceptions.InvalidAddressException:
            return False

        return True

    def get_pid(self) -> Union[int, interfaces.renderers.BaseAbsentValue]:
        """Return the pid of the process, if any."""
        if (
            self.State.description != "SERVICE_RUNNING"
            or "PROCESS" not in self.get_type()
        ):
            return renderers.NotApplicableValue()

        try:
            return self.ServiceProcess.ProcessId
        except exceptions.InvalidAddressException:
            return renderers.UnreadableValue()

    def get_binary(self) -> Union[str, interfaces.renderers.BaseAbsentValue]:
        """Returns the binary associated with the service."""
        if self.State.description != "SERVICE_RUNNING":
            return renderers.NotApplicableValue()

        # depending on whether the service is for a process
        # or kernel driver, the binary path is stored differently
        try:
            if "PROCESS" in self.get_type():
                return self.ServiceProcess.BinaryPath.dereference().cast(
                    "string", encoding="utf-16", errors="replace", max_length=512
                )
            else:
                return self.DriverName.dereference().cast(
                    "string", encoding="utf-16", errors="replace", max_length=512
                )
        except exceptions.InvalidAddressException:
            return renderers.UnreadableValue()

    def get_name(self) -> Union[str, interfaces.renderers.BaseAbsentValue]:
        """Returns the service name."""
        try:
            return self.ServiceName.dereference().cast(
                "string", encoding="utf-16", errors="replace", max_length=512
            )
        except exceptions.InvalidAddressException:
            return renderers.UnreadableValue()

    def get_display(self) -> Union[str, interfaces.renderers.BaseAbsentValue]:
        """Returns the service display."""
        try:
            return self.DisplayName.dereference().cast(
                "string", encoding="utf-16", errors="replace", max_length=512
            )
        except exceptions.InvalidAddressException:
            return renderers.UnreadableValue()

    def get_type(self) -> str:
        """Returns the binary types."""

        SERVICE_TYPE_FLAGS = {
            "SERVICE_KERNEL_DRIVER": 1,
            "SERVICE_FILE_SYSTEM_DRIVER": 2,
            "SERVICE_ADAPTOR": 4,
            "SERVICE_RECOGNIZER_DRIVER": 8,
            "SERVICE_WIN32_OWN_PROCESS": 16,
            "SERVICE_WIN32_SHARE_PROCESS": 32,
            "SERVICE_INTERACTIVE_PROCESS": 256,
        }

        type_flags = wrappers.Flags(choices=SERVICE_TYPE_FLAGS)
        return "|".join(type_flags(self.Type))

    def traverse(self):
        """Generator that enumerates other services."""

        try:
            if hasattr(self, "PrevEntry"):
                yield self
                # make sure we dereference these pointers, or the
                # is_valid() checks will apply to the pointer and
                # not the _SERVICE_RECORD object as intended.
                rec = self.PrevEntry
                while rec and rec.is_valid():
                    yield rec
                    rec = rec.PrevEntry
            else:
                rec = self
                while rec and rec.is_valid():
                    yield rec
                    rec = rec.ServiceList.Blink.dereference()
        except exceptions.InvalidAddressException:
            return None


class SERVICE_HEADER(objects.StructType):
    """A service header structure."""

    def is_valid(self) -> bool:
        """Determine if the structure is valid."""
        try:
            return self.ServiceRecord.is_valid()
        except exceptions.InvalidAddressException:
            return False


class_types = {"_SERVICE_RECORD": SERVICE_RECORD, "_SERVICE_HEADER": SERVICE_HEADER}



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import contextlib
import logging
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

from volatility3.framework import constants, exceptions, interfaces, objects
from volatility3.framework.configuration import requirements
from volatility3.framework.layers import linear
from volatility3.framework.symbols import intermed
from volatility3.framework.symbols.windows import extensions

vollog = logging.getLogger(__name__)


class RegistryException(exceptions.LayerException):
    """Base Registry Exception class for catching Registry layer errors."""


class RegistryFormatException(RegistryException):
    """Thrown when an error occurs with the underlying Registry file format."""


class RegistryInvalidIndex(RegistryException):
    """Thrown when an index that doesn't exist or can't be found occurs."""


class RegistryHive(linear.LinearlyMappedLayer):
    def __init__(
        self,
        context: interfaces.context.ContextInterface,
        config_path: str,
        name: str,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> None:
        super().__init__(
            context=context, config_path=config_path, name=name, metadata=metadata
        )

        self._base_layer = self.config["base_layer"]
        self._hive_offset = self.config["hive_offset"]
        self._table_name = self.config["nt_symbols"]
        self._page_size = 1 << 12

        self._reg_table_name = intermed.IntermediateSymbolTable.create(
            context, self._config_path, "windows", "registry"
        )

        cmhive = self.context.object(
            self._table_name + constants.BANG + "_CMHIVE",
            self._base_layer,
            self._hive_offset,
        )
        self._cmhive_name = cmhive.get_name()
        self.hive = cmhive.Hive

        # TODO: Check the checksum
        if self.hive.Signature != 0xBEE0BEE0:
            raise RegistryFormatException(
                self.name,
                f"Registry hive at {self._hive_offset} does not have a valid signature",
            )

        # Win10 17063 introduced the Registry process to map most hives.  Check
        # if it exists and update RegistryHive._base_layer
        try:
            registry_proc = self._find_registry_process()
            if registry_proc:
                self._base_layer = registry_proc.add_process_layer()
        except ValueError:
            vollog.log(
                constants.LOGLEVEL_VVVV,
                "Error walking process list, results may not be valid.",
            )

        self._base_block = self.hive.BaseBlock.dereference()

        self._minaddr = 0
        try:
            self._hive_maxaddr_non_volatile = self.hive.Storage[0].Length
            self._hive_maxaddr_volatile = self.hive.Storage[1].Length
            self._maxaddr = 0x80000000 | self._hive_maxaddr_volatile
            vollog.log(
                constants.LOGLEVEL_VVVV,
                f"Setting hive {self.name} max address to {hex(self._maxaddr)}",
            )
        except exceptions.InvalidAddressException:
            self._hive_maxaddr_non_volatile = 0x7FFFFFFF
            self._hive_maxaddr_volatile = 0x7FFFFFFF
            self._maxaddr = 0x80000000 | self._hive_maxaddr_volatile
            vollog.log(
                constants.LOGLEVEL_VVVV,
                f"Exception when setting hive {self.name} max address, using {hex(self._maxaddr)}",
            )

    def _find_registry_process(self) -> Optional[extensions.EPROCESS]:
        """Walk the active process list and return the Registry process if it exists. Duplicates
        PsList.list_processes() since pulling in the plugin causes problems.

        Returns:
            The Registry EPROCESS object if it exists, or None
        """

        kernel = self.context.modules.get(self.config["kernel_module_name"])

        if not kernel or not kernel.offset:
            raise ValueError(
                "Intel layer does not have an associated kernel virtual offset, failing"
            )

        ps_aph_offset = kernel.get_symbol("PsActiveProcessHead").address
        list_entry = kernel.object(object_type="_LIST_ENTRY", offset=ps_aph_offset)
        reloff = kernel.get_type("_EPROCESS").relative_child_offset(
            "ActiveProcessLinks"
        )
        eproc = kernel.object(
            object_type="_EPROCESS",
            offset=list_entry.vol.offset - reloff,
            absolute=True,
        )

        for proc in eproc.ActiveProcessLinks:
            proc_name = proc.ImageFileName.cast(
                "string", max_length=proc.ImageFileName.vol.count, errors="replace"
            )
            if proc_name == "Registry" and proc.InheritedFromUniqueProcessId == 4:
                return proc

        return None

    def _get_hive_maxaddr(self, volatile):
        return (
            self._hive_maxaddr_volatile if volatile else self._hive_maxaddr_non_volatile
        )

    def get_name(self) -> str:
        return self._cmhive_name or "[NONAME]"

    @property
    def hive_offset(self) -> int:
        return self._hive_offset

    @property
    def address_mask(self) -> int:
        """Return a mask that allows for the volatile bit to be set."""
        return super().address_mask | 0x80000000

    @property
    def root_cell_offset(self) -> int:
        """Returns the offset for the root cell in this hive."""
        with contextlib.suppress(exceptions.InvalidAddressException):
            if (
                self._base_block.Signature.cast(
                    "string", max_length=4, encoding="latin-1"
                )
                == "regf"
            ):
                return self._base_block.RootCell
        return 0x20

    def get_cell(self, cell_offset: int) -> "objects.StructType":
        """Returns the appropriate Cell value for a cell offset."""
        # This would be an _HCELL containing CELL_DATA, but to save time we skip the size of the HCELL
        cell = self._context.object(
            object_type=self._table_name + constants.BANG + "_CELL_DATA",
            offset=cell_offset + 4,
            layer_name=self.name,
        )
        return cell

    def get_node(self, cell_offset: int) -> "objects.StructType":
        """Returns the appropriate Node, interpreted from the Cell based on its
        Signature."""
        cell = self.get_cell(cell_offset)
        try:
            signature = cell.cast("string", max_length=2, encoding="latin-1")
        except (RegistryException, exceptions.InvalidAddressException):
            vollog.debug(
                f"Failed to get cell signature for cell (0x{cell.vol.offset:x})"
            )
            return cell
        if signature == "nk":
            return cell.u.KeyNode
        elif signature == "sk":
            return cell.u.KeySecurity
        elif signature == "vk":
            return cell.u.KeyValue
        elif signature == "db":
            # Big Data
            return cell.u.ValueData
        elif signature == "lf" or signature == "lh" or signature == "ri":
            # Fast Leaf, Hash Leaf, Index Root
            return cell.u.KeyIndex
        else:
            # It doesn't matter that we use KeyNode, we're just after the first two bytes
            vollog.debug(
                f"Unknown Signature {signature} (0x{cell.u.KeyNode.Signature:x}) at offset {cell_offset}"
            )
            return cell

    def get_key(
        self, key: str, return_list: bool = False
    ) -> Union[List[objects.StructType], objects.StructType]:
        """Gets a specific registry key by key path.

        return_list specifies whether the return result will be a single
        node (default) or a list of nodes from root to the current node
        (if return_list is true).

        Raises RegistryFormatException if an invalid structure is encountered
        Raises KeyError if the key is not found
        """
        root_node = self.get_node(self.root_cell_offset)
        if not root_node.vol.type_name.endswith(constants.BANG + "_CM_KEY_NODE"):
            raise RegistryFormatException(
                self.name,
                f"Encountered {root_node.vol.type_name} instead of _CM_KEY_NODE",
            )
        node_key = [root_node]
        if key.endswith("\\"):
            key = key[:-1]
        key_array = key.split("\\")
        found_key: List[str] = []
        while key_array and node_key:
            subkeys = node_key[-1].get_subkeys()
            for subkey in subkeys:
                # registry keys are not case sensitive so compare likewise
                # https://learn.microsoft.com/en-us/windows/win32/sysinfo/structure-of-the-registry
                if subkey.get_name().casefold() == key_array[0].casefold():
                    node_key = node_key + [subkey]
                    found_key, key_array = found_key + [key_array[0]], key_array[1:]
                    break
            else:
                node_key = []
        if not node_key:
            raise KeyError(
                "Key {} not found under {}".format(key_array[0], "\\".join(found_key))
            )
        if return_list:
            return node_key
        return node_key[-1]

    def visit_nodes(
        self,
        visitor: Callable[[objects.StructType], None],
        node: Optional[objects.StructType] = None,
    ) -> None:
        """Applies a callable (visitor) to all nodes within the registry tree
        from a given node."""
        if not node:
            node = self.get_node(self.root_cell_offset)
        visitor(node)
        for node in node.get_subkeys():
            self.visit_nodes(visitor, node)

    @staticmethod
    def _mask(value: int, high_bit: int, low_bit: int) -> int:
        """Returns the bits of a value between highbit and lowbit inclusive."""
        high_mask = (2 ** (high_bit + 1)) - 1
        low_mask = (2**low_bit) - 1
        mask = high_mask ^ low_mask
        # print(high_bit, low_bit, bin(mask), bin(value))
        return value & mask

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.IntRequirement(
                name="hive_offset",
                description="Offset within the base layer at which the hive lives",
                default=0,
                optional=False,
            ),
            requirements.SymbolTableRequirement(
                name="nt_symbols", description="Windows kernel symbols"
            ),
            requirements.TranslationLayerRequirement(
                name="base_layer",
                description="Layer in which the registry hive lives",
                optional=False,
            ),
        ]

    def _translate(self, offset: int) -> int:
        """Translates a single cell index to a cell memory offset and the
        suboffset within it."""

        # Ignore the volatile bit when determining maxaddr validity
        volatile = self._mask(offset, 31, 31) >> 31
        if offset & 0x7FFFFFFF > self._get_hive_maxaddr(volatile):
            vollog.log(
                constants.LOGLEVEL_VVV,
                "Layer {} couldn't translate offset {}, greater than {} in {} store of {}".format(
                    self.name,
                    hex(offset & 0x7FFFFFFF),
                    hex(self._get_hive_maxaddr(volatile)),
                    "volatile" if volatile else "non-volatile",
                    self.get_name(),
                ),
            )
            raise RegistryInvalidIndex(
                self.name, "Mapping request for value greater than maxaddr"
            )

        storage = self.hive.Storage[volatile]
        dir_index = self._mask(offset, 30, 21) >> 21
        table_index = self._mask(offset, 20, 12) >> 12
        suboffset = self._mask(offset, 11, 0) >> 0

        table = storage.Map.Directory[dir_index]
        entry = table.Table[table_index]
        return entry.get_block_offset() + suboffset

    def mapping(
        self, offset: int, length: int, ignore_errors: bool = False
    ) -> Iterable[Tuple[int, int, int, int, str]]:
        if length < 0:
            raise ValueError("Mapping length of RegistryHive must be positive or zero")

        # Return the translated offset without checking bounds within the HBIN.  The check runs into
        # issues when pages are swapped on large HBINs, and did not seem to find any errors on single page
        # HBINs while dramatically slowing performance.
        response = []
        current_offset = offset
        remaining_length = length
        chunk_size = self._page_size - (offset & (self._page_size - 1))
        while remaining_length > 0:
            chunk_size = min(chunk_size, remaining_length, self._page_size)
            try:
                translated_offset = self._translate(current_offset)
                response.append(
                    (
                        current_offset,
                        chunk_size,
                        translated_offset,
                        chunk_size,
                        self._base_layer,
                    )
                )
            except exceptions.LayerException:
                if not ignore_errors:
                    raise
            current_offset += chunk_size
            remaining_length -= chunk_size
            chunk_size = self._page_size
        return response

    @property
    def dependencies(self) -> List[str]:
        """Returns a list of layer names that this layer translates onto."""
        return [self.config["base_layer"]]

    def is_valid(self, offset: int, length: int = 1) -> bool:
        """Returns a boolean based on whether the offset is valid or not."""
        with contextlib.suppress(exceptions.InvalidAddressException):
            # Pass this to the lower layers for now
            return all(
                self.context.layers[layer].is_valid(offset, length)
                for (_, _, offset, length, layer) in self.mapping(offset, length)
            )
        return False

    @property
    def minimum_address(self) -> int:
        return self._minaddr

    @property
    def maximum_address(self) -> int:
        return self._maxaddr



from volatility3.framework.symbols import intermed
from volatility3.framework.symbols.linux.extensions import network
from volatility3.framework.interfaces import configuration


class NetSymbols(configuration.VersionableInterface):
    _version = (1, 0, 0)

    @classmethod
    def apply(cls, symbol_table: intermed.IntermediateSymbolTable):
        # Network
        symbol_table.set_type_class("net", network.net)
        symbol_table.set_type_class("net_device", network.net_device)
        symbol_table.set_type_class("in_device", network.in_device)
        symbol_table.set_type_class("in_ifaddr", network.in_ifaddr)
        symbol_table.set_type_class("inet6_dev", network.inet6_dev)
        symbol_table.set_type_class("inet6_ifaddr", network.inet6_ifaddr)
        symbol_table.set_type_class("socket", network.socket)
        symbol_table.set_type_class("sock", network.sock)
        symbol_table.set_type_class("inet_sock", network.inet_sock)
        symbol_table.set_type_class("unix_sock", network.unix_sock)
        # Might not exist in older kernels or the current symbols
        symbol_table.optional_set_type_class("netlink_sock", network.netlink_sock)
        symbol_table.optional_set_type_class("vsock_sock", network.vsock_sock)
        symbol_table.optional_set_type_class("packet_sock", network.packet_sock)
        symbol_table.optional_set_type_class("bt_sock", network.bt_sock)
        symbol_table.optional_set_type_class("xdp_sock", network.xdp_sock)



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
"""Volatility 3 Windows Constants.

Windows-specific values that aren't found in debug symbols
"""

KERNEL_MODULE_NAMES = ["ntkrnlmp", "ntkrnlpa", "ntkrpamp", "ntoskrnl"]
"""The list of names that kernel modules can have within the windows OS"""

PE_MAX_EXTRACTION_SIZE = 1024 * 1024 * 256

"""
The following constants represent the layout of the Low Stub which exists only on x64 machines with no virtualization/emulation,
responsible for transitioning from Real Mode(16 bit) to Protected Mode(32 bit) and Long Mode(64 bit) on boot/return from sleep.
Contains offsets to fields and structures within the undocumented structure _PROCESSOR_START_BLOCK.
Here's a reference: https://github.com/mic101/windows/blob/master/WRK-v1.2/base/ntos/inc/amd64.h#L3334
"""
# Expected signature for validation, constructed from:
# PROCESSOR_START_BLOCK->Jmp->OpCode | PROCESSOR_START_BLOCK->Jmp->Offset | PROCESSOR_START_BLOCK->CompletionFlag
JMP_AND_COMPLETION_SIGNATURE = 0x00000001000600E9

# Address of LmTarget (Long Mode target)
PROCESSOR_START_BLOCK_LM_TARGET_OFFSET = (
    0x70  # PROCESSOR_START_BLOCK->LmTarget, PVOID 8 bytes
)

# CR3 register within structures describing initial processor state to be started
PROCESSOR_START_BLOCK_CR3_OFFSET = 0xA0  # PROCESSOR_START_BLOCK->ProcessorState->SpecialRegisters->Cr3, ULONG64 8 bytes

MAX_PID = 0xFFFFFFFC



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
from typing import Generator, Tuple

from volatility3.framework import constants, interfaces, objects
from volatility3.framework.renderers import conversion

vollog = logging.getLogger(__name__)


class IMAGE_DOS_HEADER(objects.StructType):
    def get_nt_header(self) -> interfaces.objects.ObjectInterface:
        """Carve out the NT header from this DOS header. This reflects on the
        PE file's Machine type to create a 32- or 64-bit NT header structure.

        Returns:
            <_IMAGE_NT_HEADERS> or <_IMAGE_NT_HEADERS64> instance
        """

        if self.e_magic != 0x5A4D:
            raise ValueError(
                f"e_magic {self.e_magic:04X} is not a valid DOS signature."
            )

        layer_name = self.vol.layer_name
        symbol_table_name = self.get_symbol_table_name()

        nt_header = self._context.object(
            symbol_table_name + constants.BANG + "_IMAGE_NT_HEADERS",
            layer_name=layer_name,
            offset=self.vol.offset + self.e_lfanew,
        )

        if nt_header.Signature != 0x4550:
            raise ValueError(
                f"NT header signature {nt_header.Signature:04X} is not a valid"
            )

        # this checks if we need a PE32+ header
        if nt_header.FileHeader.Machine == 34404:
            nt_header = nt_header.cast("_IMAGE_NT_HEADERS64")

        return nt_header

    def replace_header_field(
        self,
        sect: interfaces.objects.ObjectInterface,
        header: bytes,
        item: interfaces.objects.ObjectInterface,
        value: int,
    ) -> bytes:
        """Replaces a member in an _IMAGE_SECTION_HEADER structure.

        Args:
            sect: the section instance
            header: raw data for the section
            item: the member of the section to replace
            value: new value for the member

        Returns:
            The raw data with the replaced header field
        """

        member_size = self._context.symbol_space.get_type(item.vol.type_name).size
        start = item.vol.offset - sect.vol.offset
        newval = objects.convert_value_to_data(value, int, item.vol.data_format)
        result = header[:start] + newval + header[start + member_size :]
        return result

    def fix_image_base(
        self, raw_data: bytes, nt_header: interfaces.objects.ObjectInterface
    ) -> bytes:
        """Fix the _OPTIONAL_HEADER.ImageBase value (which is either an
        unsigned long for 32-bit PE's or unsigned long long for 64-bit PE's) to
        match the address where the PE file was carved out of memory.

        Args:
            raw_data: a bytes object of the PE's data
            nt_header: <_IMAGE_NT_HEADERS> or <_IMAGE_NT_HEADERS64> instance

        Returns:
             <bytes> patched with the correct address
        """

        image_base_offset = (
            nt_header.OptionalHeader.ImageBase.vol.offset - self.vol.offset
        )
        image_base_type = nt_header.OptionalHeader.ImageBase.vol.type_name
        member_size = self._context.symbol_space.get_type(image_base_type).size
        try:
            newval = objects.convert_value_to_data(
                self.vol.offset, int, nt_header.OptionalHeader.ImageBase.vol.data_format
            )
            new_pe = (
                raw_data[:image_base_offset]
                + newval
                + raw_data[image_base_offset + member_size :]
            )
        except OverflowError:
            vollog.warning(
                f"Volatility was unable to fix the image base for the PE file at base address {self.vol.offset:#x}. "
                "This will cause issues with many static analysis tools if you do not inform the "
                "tool of the in-memory load address."
            )
            new_pe = raw_data

        return new_pe

    def reconstruct(self) -> Generator[Tuple[int, bytes], None, None]:
        """This method generates the content necessary to reconstruct a PE file
        from memory. It preserves slack space (similar to the old --memory) and
        automatically fixes the ImageBase in the output PE file.

        Returns:
            <tuple> of (<int> offset, <bytes> data)
        """

        nt_header = self.get_nt_header()

        layer_name = self.vol.layer_name
        symbol_table_name = self.get_symbol_table_name()

        section_alignment = nt_header.OptionalHeader.SectionAlignment

        sect_header_size = self._context.symbol_space.get_type(
            symbol_table_name + constants.BANG + "_IMAGE_SECTION_HEADER"
        ).size

        size_of_image = nt_header.OptionalHeader.SizeOfImage

        # no legitimate PE is going to be larger than this
        if size_of_image > constants.windows.PE_MAX_EXTRACTION_SIZE:
            raise ValueError(f"The claimed SizeOfImage is too large: {size_of_image}")

        read_layer = self._context.layers[layer_name]

        raw_data = read_layer.read(
            self.vol.offset, nt_header.OptionalHeader.SizeOfImage, pad=True
        )

        # fix the PE image base before yielding the initial view of the data
        fixed_data = self.fix_image_base(raw_data, nt_header)
        yield 0, fixed_data

        start_addr = nt_header.FileHeader.SizeOfOptionalHeader + (
            nt_header.OptionalHeader.vol.offset - self.vol.offset
        )

        counter = 0
        for sect in nt_header.get_sections():
            if sect.VirtualAddress > size_of_image:
                raise ValueError(
                    f"Section VirtualAddress is too large: {sect.VirtualAddress}"
                )

            if sect.Misc.VirtualSize > size_of_image:
                raise ValueError(
                    f"Section VirtualSize is too large: {sect.Misc.VirtualSize}"
                )

            if sect.SizeOfRawData > size_of_image:
                raise ValueError(
                    f"Section SizeOfRawData is too large: {sect.SizeOfRawData}"
                )

            if sect is not None:
                # It doesn't matter if this is too big, because it'll get overwritten by the later layers
                sect_size = conversion.round(
                    sect.Misc.VirtualSize, section_alignment, up=True
                )
                sectheader = read_layer.read(sect.vol.offset, sect_header_size)
                sectheader = self.replace_header_field(
                    sect, sectheader, sect.PointerToRawData, sect.VirtualAddress
                )
                sectheader = self.replace_header_field(
                    sect, sectheader, sect.SizeOfRawData, sect_size
                )
                sectheader = self.replace_header_field(
                    sect, sectheader, sect.Misc.VirtualSize, sect_size
                )

                offset = start_addr + (counter * sect_header_size)
                yield offset, sectheader
                counter += 1


class IMAGE_NT_HEADERS(objects.StructType):
    def get_sections(self) -> Generator[interfaces.objects.ObjectInterface, None, None]:
        """Iterate through the section headers for this PE file.

        Yields:
            <_IMAGE_SECTION_HEADER> objects
        """
        layer_name = self.vol.layer_name
        symbol_table_name = self.get_symbol_table_name()

        sect_header_size = self._context.symbol_space.get_type(
            symbol_table_name + constants.BANG + "_IMAGE_SECTION_HEADER"
        ).size
        start_addr = (
            self.FileHeader.SizeOfOptionalHeader + self.OptionalHeader.vol.offset
        )

        for i in range(self.FileHeader.NumberOfSections):
            sect_addr = start_addr + (i * sect_header_size)
            yield self._context.object(
                symbol_table_name + constants.BANG + "_IMAGE_SECTION_HEADER",
                offset=sect_addr,
                layer_name=layer_name,
            )


class_types = {
    "_IMAGE_DOS_HEADER": IMAGE_DOS_HEADER,
    # the 32- and 64-bit extensions behave the same way, but the underlying structure is different
    "_IMAGE_NT_HEADERS": IMAGE_NT_HEADERS,
    "_IMAGE_NT_HEADERS64": IMAGE_NT_HEADERS,
}



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
from typing import Callable, List, Generator, Iterable, Type, Optional, Tuple

from volatility3.framework import renderers, interfaces, exceptions
from volatility3.framework.configuration import requirements
from volatility3.framework.objects import utility
from volatility3.framework.renderers import format_hints
from volatility3.plugins.windows import pslist

vollog = logging.getLogger(__name__)

# these are from WinNT.h
winnt_protections = {
    "PAGE_NOACCESS": 0x01,
    "PAGE_READONLY": 0x02,
    "PAGE_READWRITE": 0x04,
    "PAGE_WRITECOPY": 0x08,
    "PAGE_EXECUTE": 0x10,
    "PAGE_EXECUTE_READ": 0x20,
    "PAGE_EXECUTE_READWRITE": 0x40,
    "PAGE_EXECUTE_WRITECOPY": 0x80,
    "PAGE_GUARD": 0x100,
    "PAGE_NOCACHE": 0x200,
    "PAGE_WRITECOMBINE": 0x400,
    "PAGE_TARGETS_INVALID": 0x40000000,
}


class VadInfo(interfaces.plugins.PluginInterface):
    """Lists process memory ranges."""

    _required_framework_version = (2, 4, 0)
    _version = (2, 0, 1)
    MAXSIZE_DEFAULT = 1024 * 1024 * 1024  # 1 Gb

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._protect_values = None

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        # Since we're calling the plugin, make sure we have the plugin's requirements
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            # TODO: Convert this to a ListRequirement so that people can filter on sets of ranges
            requirements.IntRequirement(
                name="address",
                description="Process virtual memory address to include "
                "(all other address ranges are excluded).",
                optional=True,
            ),
            requirements.ListRequirement(
                name="pid",
                description="Filter on specific process IDs",
                element_type=int,
                optional=True,
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(3, 0, 0)
            ),
            requirements.BooleanRequirement(
                name="dump",
                description="Extract listed memory ranges",
                default=False,
                optional=True,
            ),
            requirements.IntRequirement(
                name="maxsize",
                description="Maximum size for dumped VAD sections "
                "(all the bigger sections will be ignored)",
                default=cls.MAXSIZE_DEFAULT,
                optional=True,
            ),
        ]

    @classmethod
    def protect_values(
        cls,
        context: interfaces.context.ContextInterface,
        layer_name: str,
        symbol_table: str,
    ) -> Iterable[int]:
        """Look up the array of memory protection constants from the memory
        sample. These don't change often, but if they do in the future, then
        finding them dynamically versus hard-coding here will ensure we parse
        them properly.

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            layer_name: The name of the layer on which to operate
            symbol_table: The name of the table containing the kernel symbols
        """

        kvo = context.layers[layer_name].config.get("kernel_virtual_offset", None)
        if not kvo:
            raise ValueError(
                "Intel layer does not have an associated kernel virtual offset, failing"
            )
        ntkrnlmp = context.module(symbol_table, layer_name=layer_name, offset=kvo)
        addr = ntkrnlmp.get_symbol("MmProtectToValue").address
        values = ntkrnlmp.object(
            object_type="array", offset=addr, subtype=ntkrnlmp.get_type("int"), count=32
        )
        return values  # type: ignore

    @classmethod
    def list_vads(
        cls,
        proc: interfaces.objects.ObjectInterface,
        filter_func: Callable[[interfaces.objects.ObjectInterface], bool] = lambda _: (
            False
        ),
    ) -> Generator[interfaces.objects.ObjectInterface, None, None]:
        """Lists the Virtual Address Descriptors of a specific process.

        Args:
            proc: _EPROCESS object from which to list the VADs
            filter_func: Function to take a virtual address descriptor value and return True if it should be filtered out

        Returns:
            A list of virtual address descriptors based on the process and filtered based on the filter function
        """
        for vad in proc.get_vad_root().traverse():
            if not filter_func(vad):
                yield vad

    @classmethod
    def vad_dump(
        cls,
        context: interfaces.context.ContextInterface,
        proc: interfaces.objects.ObjectInterface,
        vad: interfaces.objects.ObjectInterface,
        open_method: Type[interfaces.plugins.FileHandlerInterface],
        maxsize: int = MAXSIZE_DEFAULT,
    ) -> Optional[interfaces.plugins.FileHandlerInterface]:
        """Extracts the complete data for Vad as a FileInterface.

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            proc: an _EPROCESS instance
            vad: The suspected VAD to extract (ObjectInterface)
            open_method: class to provide context manager for opening the file
            maxsize: Max size of VAD section (default MAXSIZE_DEFAULT)

        Returns:
            An open FileInterface object containing the complete data for the process or None in the case of failure
        """

        try:
            vad_start = vad.get_start()
            vad_end = vad.get_end()
        except AttributeError:
            vollog.debug("Unable to find the starting/ending VPN member")
            return None

        if 0 < maxsize < vad.get_size():
            vollog.debug(
                f"Skip VAD dump {vad_start:#x}-{vad_end:#x} due to maxsize limit"
            )
            return None

        proc_id = "Unknown"
        try:
            proc_id = proc.UniqueProcessId
            proc_layer_name = proc.add_process_layer()
        except exceptions.InvalidAddressException as excp:
            vollog.debug(
                f"Process {proc_id}: invalid address {excp.invalid_address} in layer {excp.layer_name}"
            )
            return None

        proc_layer = context.layers[proc_layer_name]
        file_name = f"pid.{proc_id}.vad.{vad_start:#x}-{vad_end:#x}.dmp"
        try:
            file_handle = open_method(file_name)
            chunk_size = 1024 * 1024 * 10
            offset = vad_start
            vad_size = vad.get_size()
            while offset < vad_start + vad_size:
                to_read = min(chunk_size, vad_start + vad_size - offset)
                data = proc_layer.read(offset, to_read, pad=True)
                if not data:
                    break
                file_handle.write(data)
                offset += to_read

        except Exception as excp:
            vollog.debug(f"Unable to dump VAD {file_name}: {excp}")
            return None

        return file_handle

    def _generator(
        self, procs: List[interfaces.objects.ObjectInterface]
    ) -> Generator[
        Tuple[
            int,
            Tuple[
                int,
                str,
                format_hints.Hex,
                format_hints.Hex,
                format_hints.Hex,
                str,
                str,
                int,
                int,
                format_hints.Hex,
                str,
                str,
            ],
        ],
        None,
        None,
    ]:
        kernel = self.context.modules[self.config["kernel"]]
        kernel_layer = self.context.layers[kernel.layer_name]

        def passthrough(x: interfaces.objects.ObjectInterface) -> bool:
            return False

        filter_func = passthrough
        if self.config.get("address", None) is not None:

            def filter_function(x: interfaces.objects.ObjectInterface) -> bool:
                return not (x.get_start() <= self.config["address"] <= x.get_end())

            filter_func = filter_function

        for proc in procs:
            process_name = utility.array_to_string(proc.ImageFileName)

            for vad in self.list_vads(proc, filter_func=filter_func):
                file_output = "Disabled"
                if self.config["dump"]:
                    file_handle = self.vad_dump(
                        self.context, proc, vad, self.open, self.config["maxsize"]
                    )
                    file_output = "Error outputting file"
                    if file_handle:
                        file_handle.close()
                        file_output = file_handle.preferred_filename

                yield (
                    0,
                    (
                        proc.UniqueProcessId,
                        process_name,
                        format_hints.Hex(kernel_layer.canonicalize(vad.vol.offset)),
                        format_hints.Hex(vad.get_start()),
                        format_hints.Hex(vad.get_end()),
                        vad.get_tag(),
                        vad.get_protection(
                            self.protect_values(
                                self.context,
                                kernel.layer_name,
                                kernel.symbol_table_name,
                            ),
                            winnt_protections,
                        ),
                        vad.get_commit_charge(),
                        vad.get_private_memory(),
                        format_hints.Hex(vad.get_parent()),
                        vad.get_file_name(),
                        file_output,
                    ),
                )

    def run(self) -> renderers.TreeGrid:
        filter_func = pslist.PsList.create_pid_filter(self.config.get("pid", None))

        return renderers.TreeGrid(
            [
                ("PID", int),
                ("Process", str),
                ("Offset", format_hints.Hex),
                ("Start VPN", format_hints.Hex),
                ("End VPN", format_hints.Hex),
                ("Tag", str),
                ("Protection", str),
                ("CommitCharge", int),
                ("PrivateMemory", int),
                ("Parent", format_hints.Hex),
                ("File", str),
                ("File output", str),
            ],
            self._generator(
                pslist.PsList.list_processes(
                    context=self.context,
                    kernel_module_name=self.config["kernel"],
                    filter_func=filter_func,
                )
            ),
        )



# This file is Copyright 2025 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

# Public researches: https://i.blackhat.com/USA21/Wednesday-Handouts/us-21-Fixing-A-Memory-Forensics-Blind-Spot-Linux-Kernel-Tracing-wp.pdf

import logging
from typing import List, Generator
from enum import Enum
from dataclasses import dataclass

import volatility3.framework.symbols.linux.utilities.modules as linux_utilities_modules
from volatility3.framework import constants, exceptions, interfaces, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.renderers import format_hints
from volatility3.framework.constants import architectures

vollog = logging.getLogger(__name__)


# https://docs.python.org/3.13/library/enum.html#enum.IntFlag
class FtraceOpsFlags(Enum):
    """Denote the state of an ftrace_ops struct.
    Based on https://elixir.bootlin.com/linux/v6.13-rc3/source/include/linux/ftrace.h#L255.
    """

    FTRACE_OPS_FL_ENABLED = 1 << 0
    FTRACE_OPS_FL_DYNAMIC = 1 << 1
    FTRACE_OPS_FL_SAVE_REGS = 1 << 2
    FTRACE_OPS_FL_SAVE_REGS_IF_SUPPORTED = 1 << 3
    FTRACE_OPS_FL_RECURSION = 1 << 4
    FTRACE_OPS_FL_STUB = 1 << 5
    FTRACE_OPS_FL_INITIALIZED = 1 << 6
    FTRACE_OPS_FL_DELETED = 1 << 7
    FTRACE_OPS_FL_ADDING = 1 << 8
    FTRACE_OPS_FL_REMOVING = 1 << 9
    FTRACE_OPS_FL_MODIFYING = 1 << 10
    FTRACE_OPS_FL_ALLOC_TRAMP = 1 << 11
    FTRACE_OPS_FL_IPMODIFY = 1 << 12
    FTRACE_OPS_FL_PID = 1 << 13
    FTRACE_OPS_FL_RCU = 1 << 14
    FTRACE_OPS_FL_TRACE_ARRAY = 1 << 15
    FTRACE_OPS_FL_PERMANENT = 1 << 16
    FTRACE_OPS_FL_DIRECT = 1 << 17
    FTRACE_OPS_FL_SUBOP = 1 << 18


@dataclass
class ParsedFtraceOps:
    """Parsed ftrace_ops struct representation, containing a selection of forensics valuable
    information."""

    ftrace_ops_offset: int
    callback_symbol: str
    callback_address: int
    hooked_symbols: str
    module_name: str
    module_address: int
    flags: str


class CheckFtrace(interfaces.plugins.PluginInterface):
    """Detect ftrace hooking

    Investigate the ftrace infrastructure to uncover kernel attached callbacks, which can be leveraged
    to hook kernel functions and modify their behaviour."""

    _version = (4, 0, 0)
    _required_framework_version = (2, 19, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Linux kernel",
                architectures=architectures.LINUX_ARCHS,
            ),
            requirements.VersionRequirement(
                name="linux_utilities_modules",
                component=linux_utilities_modules.Modules,
                version=(3, 0, 0),
            ),
            requirements.VersionRequirement(
                name="linux_utilities_module_gatherers",
                component=linux_utilities_modules.ModuleGatherers,
                version=(1, 0, 0),
            ),
            requirements.BooleanRequirement(
                name="show_ftrace_flags",
                description="Show ftrace flags associated with an ftrace_ops struct",
                optional=True,
                default=False,
            ),
        ]

    @classmethod
    def extract_hash_table_filters(
        cls,
        ftrace_ops: interfaces.objects.ObjectInterface,
    ) -> Generator[interfaces.objects.ObjectInterface, None, None]:
        """Wrap the process of walking to every ftrace_func_entry of an ftrace_ops.
        Those are stored in a hash table of filters that indicates the addresses hooked.

        Args:
            ftrace_ops: The ftrace_ops struct to walk through

        Return, None, None:
            An iterable of ftrace_func_entry structs
        """

        if hasattr(ftrace_ops, "func_hash"):
            ftrace_hash = ftrace_ops.func_hash.filter_hash
        else:
            ftrace_hash = ftrace_ops.filter_hash

        try:
            current_bucket_ptr = ftrace_hash.buckets.first
        except exceptions.InvalidAddressException:
            vollog.log(
                constants.LOGLEVEL_VV,
                f"ftrace_func_entry list of ftrace_ops@{ftrace_ops.vol.offset:#x} is empty/invalid. Skipping it...",
            )
            return

        while current_bucket_ptr.is_readable():
            yield current_bucket_ptr.dereference().cast("ftrace_func_entry")
            current_bucket_ptr = current_bucket_ptr.next

    @classmethod
    def parse_ftrace_ops(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
        known_modules: List[linux_utilities_modules.ModuleInfo],
        ftrace_ops: interfaces.objects.ObjectInterface,
    ) -> Generator[ParsedFtraceOps, None, None]:
        """Parse an ftrace_ops struct to highlight ftrace kernel hooking.
        Iterates over embedded ftrace_func_entry entries, which point to hooked memory areas.

        Args:
            known_modules: A dict of known modules, used to locate callbacks origin. Typically obtained through run_modules_scanners().
            ftrace_ops: The ftrace_ops struct to parse

        Yields:
            An iterable of ParsedFtraceOps dataclasses, containing a selection of useful fields (callback, hook, module) related to an ftrace_ops struct
        """
        kernel = context.modules[kernel_module_name]
        callback = ftrace_ops.func

        mod_info, callback_symbol = (
            linux_utilities_modules.Modules.module_lookup_by_address(
                context,
                kernel_module_name,
                known_modules,
                callback,
            )
        )

        if mod_info:
            module_address = mod_info.start
            module_name = mod_info.name
        else:
            callback_symbol = module_address = module_name = None

            vollog.debug(
                f"Could not determine ftrace_ops@{ftrace_ops.vol.offset:#x} callback {callback:#x} module origin.",
            )

        # Iterate over ftrace_func_entry list
        for ftrace_func_entry in cls.extract_hash_table_filters(ftrace_ops):
            hook_address = ftrace_func_entry.ip.cast("pointer")

            # Determine the symbols associated with a hook
            hooked_symbols = kernel.get_symbols_by_absolute_location(hook_address)
            hooked_symbols = ",".join(
                [
                    hooked_symbol.split(constants.BANG)[-1]
                    for hooked_symbol in hooked_symbols
                ]
            )
            formatted_ftrace_flags = ",".join(
                [flag.name for flag in FtraceOpsFlags if flag.value & ftrace_ops.flags]
            )
            yield ParsedFtraceOps(
                ftrace_ops.vol.offset,
                callback_symbol,
                callback,
                hooked_symbols,
                module_name,
                module_address,
                formatted_ftrace_flags,
            )

    @classmethod
    def iterate_ftrace_ops_list(
        cls, context: interfaces.context.ContextInterface, kernel_name: str
    ) -> Generator[interfaces.objects.ObjectInterface, None, None]:
        """Iterate over (ftrace_ops *)ftrace_ops_list.

        Returns:
            An iterable of ftrace_ops structs
        """
        kernel = context.modules[kernel_name]
        current_frace_ops_ptr = kernel.object_from_symbol("ftrace_ops_list")
        ftrace_list_end = kernel.object_from_symbol("ftrace_list_end")

        while current_frace_ops_ptr.is_readable():
            # ftrace_list_end is not considered a valid struct
            # see kernel function test_rec_ops_needs_regs
            if current_frace_ops_ptr != ftrace_list_end.vol.offset:
                yield current_frace_ops_ptr.dereference()
                current_frace_ops_ptr = current_frace_ops_ptr.next
            else:
                break

    def _generator(self):
        kernel_name = self.config["kernel"]
        kernel = self.context.modules[kernel_name]

        if not kernel.has_symbol("ftrace_ops_list"):
            vollog.error(
                'The provided symbol table does not include the "ftrace_ops_list" symbol. This means you are either analyzing an unsupported kernel version or that your symbol table is corrupted.'
            )
            return

        known_modules = linux_utilities_modules.Modules.run_modules_scanners(
            context=self.context,
            kernel_module_name=self.config["kernel"],
            caller_wanted_gatherers=linux_utilities_modules.ModuleGatherers.all_gatherers_identifier,
        )

        for ftrace_ops in self.iterate_ftrace_ops_list(self.context, kernel_name):
            for ftrace_ops_parsed in self.parse_ftrace_ops(
                self.context,
                kernel_name,
                known_modules,
                ftrace_ops,
            ):
                formatted_results = (
                    format_hints.Hex(ftrace_ops_parsed.ftrace_ops_offset),
                    ftrace_ops_parsed.callback_symbol or renderers.NotAvailableValue(),
                    format_hints.Hex(ftrace_ops_parsed.callback_address),
                    ftrace_ops_parsed.hooked_symbols or renderers.NotAvailableValue(),
                    ftrace_ops_parsed.module_name or renderers.NotAvailableValue(),
                    (
                        format_hints.Hex(ftrace_ops_parsed.module_address)
                        if ftrace_ops_parsed.module_address is not None
                        else renderers.NotAvailableValue()
                    ),
                )
                if self.config["show_ftrace_flags"]:
                    formatted_results += (ftrace_ops_parsed.flags,)
                yield (0, formatted_results)

    def run(self):
        columns = [
            ("ftrace_ops address", format_hints.Hex),
            ("Callback", str),
            ("Callback address", format_hints.Hex),
            ("Hooked symbols", str),
            ("Module", str),
            ("Module address", format_hints.Hex),
        ]

        if self.config.get("show_ftrace_flags"):
            columns.append(("Flags", str))

        return renderers.TreeGrid(
            columns,
            self._generator(),
        )



# This file is Copyright 2025 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import logging
from typing import List, Iterable

from volatility3.framework import interfaces, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.renderers import format_hints
from volatility3.plugins.windows import windowstations

vollog = logging.getLogger(__name__)


class Desktops(interfaces.plugins.PluginInterface):
    """Enumerates the Desktop instances of each Window Station"""

    _required_framework_version = (2, 0, 0)
    _version = (1, 0, 0)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.implementation = self.list_desktops

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        # Since we're calling the plugin, make sure we have the plugin's requirements
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="windowstations",
                component=windowstations.WindowStations,
                version=(1, 0, 0),
            ),
        ]

    @classmethod
    def list_desktops(
        cls,
        context: interfaces.context.ContextInterface,
        config_path: str,
        kernel_module_name: str,
    ) -> Iterable[interfaces.objects.ObjectInterface]:
        """
        Uses `scan_window_stations` to find each window station
        For each found, enumerates its desktops followed by the
        threads of each desktop.
        """
        kernel = context.modules[kernel_module_name]

        for (
            winsta,
            station_name,
            session_id,
        ) in windowstations.WindowStations.scan_window_stations(
            context, config_path, kernel_module_name
        ):
            # for each window station, walk its list of desktops
            for desktop, desktop_name in winsta.desktops(kernel.symbol_table_name):
                # for each desktop, walk its threads
                for _thread, process_name, process_pid in desktop.get_threads():
                    yield (
                        format_hints.Hex(desktop.vol.offset),
                        station_name,
                        session_id,
                        desktop_name,
                        process_name,
                        process_pid,
                    )

    def _generator(self):
        kernel_name = self.config["kernel"]

        # call the implementation for finding desktops
        # yield the information, which will include the owning window station and process
        for desktop_info in self.implementation(
            self.context, self.config_path, kernel_name
        ):
            yield 0, desktop_info

    def run(self):
        return renderers.TreeGrid(
            [
                ("Offset", format_hints.Hex),
                ("Window Station", str),
                ("Session", int),
                ("Desktop", str),
                ("Process", str),
                ("PID", int),
            ],
            self._generator(),
        )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import logging
from typing import List

from volatility3.framework import exceptions, interfaces
from volatility3.framework import renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.interfaces import plugins
from volatility3.framework.renderers import format_hints
from volatility3.framework.symbols import mac
from volatility3.plugins.mac import lsmod

vollog = logging.getLogger(__name__)


class Check_syscall(plugins.PluginInterface):
    """Check system call table for hooks."""

    _required_framework_version = (2, 0, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Kernel module for the OS",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="macutils", component=mac.MacUtilities, version=(1, 0, 0)
            ),
            requirements.VersionRequirement(
                name="lsmod", component=lsmod.Lsmod, version=(2, 0, 0)
            ),
        ]

    def _generator(self):
        kernel = self.context.modules[self.config["kernel"]]

        mods = lsmod.Lsmod.list_modules(self.context, self.config["kernel"])

        handlers = mac.MacUtilities.generate_kernel_handler_info(
            self.context, kernel.layer_name, kernel, mods
        )

        table = kernel.object_from_symbol(symbol_name="sysent")

        for i, ent in enumerate(table):
            try:
                call_addr = ent.sy_call.dereference().vol.offset
            except exceptions.InvalidAddressException:
                continue

            if not call_addr or call_addr == 0:
                continue

            module_name, symbol_name = mac.MacUtilities.lookup_module_address(
                self.context, handlers, call_addr, self.config["kernel"]
            )

            yield (
                0,
                (
                    format_hints.Hex(table.vol.offset),
                    "SysCall",
                    i,
                    format_hints.Hex(call_addr),
                    module_name,
                    symbol_name,
                ),
            )

    def run(self):
        return renderers.TreeGrid(
            [
                ("Table Address", format_hints.Hex),
                ("Table Name", str),
                ("Index", int),
                ("Handler Address", format_hints.Hex),
                ("Handler Module", str),
                ("Handler Symbol", str),
            ],
            self._generator(),
        )



# This file is Copyright 2019 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import binascii
import bz2
import datetime
import gzip
import json
import logging
import lzma
import os
from bisect import bisect
from typing import Tuple, Dict, Any, Optional, Union, List
from urllib import request, error, parse

from volatility3.framework import contexts, interfaces, constants
from volatility3.framework.layers import physical, msf, resources

vollog = logging.getLogger(__name__)

primitives = {
    0x03: ("void", {"endian": "little", "kind": "void", "signed": True, "size": 0}),
    0x08: ("HRESULT", {"endian": "little", "kind": "int", "signed": False, "size": 4}),
    0x10: ("char", {"endian": "little", "kind": "char", "signed": True, "size": 1}),
    0x20: (
        "unsigned char",
        {"endian": "little", "kind": "char", "signed": False, "size": 1},
    ),
    0x68: ("int8", {"endian": "little", "kind": "int", "signed": True, "size": 1}),
    0x69: ("uint8", {"endian": "little", "kind": "int", "signed": False, "size": 1}),
    0x70: ("char", {"endian": "little", "kind": "char", "signed": True, "size": 1}),
    0x71: ("wchar", {"endian": "little", "kind": "int", "signed": True, "size": 2}),
    # 0x7a: ("rchar16", {}),
    # 0x7b: ("rchar32", {}),
    0x11: ("short", {"endian": "little", "kind": "int", "signed": True, "size": 2}),
    0x21: (
        "unsigned short",
        {"endian": "little", "kind": "int", "signed": False, "size": 2},
    ),
    0x72: ("short", {"endian": "little", "kind": "int", "signed": True, "size": 2}),
    0x73: (
        "unsigned short",
        {"endian": "little", "kind": "int", "signed": False, "size": 2},
    ),
    0x12: ("long", {"endian": "little", "kind": "int", "signed": True, "size": 4}),
    0x22: (
        "unsigned long",
        {"endian": "little", "kind": "int", "signed": False, "size": 4},
    ),
    0x74: ("int", {"endian": "little", "kind": "int", "signed": True, "size": 4}),
    0x75: (
        "unsigned int",
        {"endian": "little", "kind": "int", "signed": False, "size": 4},
    ),
    0x13: ("long long", {"endian": "little", "kind": "int", "signed": True, "size": 8}),
    0x23: (
        "unsigned long long",
        {"endian": "little", "kind": "int", "signed": False, "size": 8},
    ),
    0x76: ("long long", {"endian": "little", "kind": "int", "signed": True, "size": 8}),
    0x77: (
        "unsigned long long",
        {"endian": "little", "kind": "int", "signed": False, "size": 8},
    ),
    0x14: ("int128", {"endian": "little", "kind": "int", "signed": True, "size": 16}),
    0x24: ("uint128", {"endian": "little", "kind": "int", "signed": False, "size": 16}),
    0x78: ("int128", {"endian": "little", "kind": "int", "signed": True, "size": 16}),
    0x79: ("uint128", {"endian": "little", "kind": "int", "signed": False, "size": 16}),
    0x46: ("f16", {"endian": "little", "kind": "float", "signed": True, "size": 2}),
    0x40: ("f32", {"endian": "little", "kind": "float", "signed": True, "size": 4}),
    0x45: ("f32pp", {"endian": "little", "kind": "float", "signed": True, "size": 4}),
    0x44: ("f48", {"endian": "little", "kind": "float", "signed": True, "size": 6}),
    0x41: ("double", {"endian": "little", "kind": "float", "signed": True, "size": 8}),
    0x42: ("f80", {"endian": "little", "kind": "float", "signed": True, "size": 10}),
    0x43: ("f128", {"endian": "little", "kind": "float", "signed": True, "size": 16}),
}

indirections = {
    0x100: (
        "pointer16",
        {"endian": "little", "kind": "int", "signed": False, "size": 2},
    ),
    0x400: (
        "pointer32",
        {"endian": "little", "kind": "int", "signed": False, "size": 4},
    ),
    0x600: (
        "pointer64",
        {"endian": "little", "kind": "int", "signed": False, "size": 8},
    ),
}


class ForwardArrayCount:
    def __init__(self, size, element_type):
        self.element_type = element_type
        self.size = size


class PdbReader:
    """Class to read Microsoft PDB files.

    This reads the various streams according to various sources as to how pdb should be read.
    These sources include:

    https://docs.rs/crate/pdb/0.5.0/source/src/
    https://github.com/moyix/pdbparse
    https://llvm.org/docs/PDB/index.html
    https://github.com/Microsoft/microsoft-pdb/

    In order to generate ISF files, we need the type stream (2), and the symbols stream (variable).
    The MultiStream Format wrapper is handled as a volatility layer, which constructs sublayers for each stream.
    The streams can then be read contiguously allowing the data to be accessed.

    Volatility's type system is strong when everything must be laid out in advance, but PDB data is reasonably dynamic,
    particularly when it comes to names.  We must therefore parse it after we've collected other information already.
    This is in comparison to something such as Construct/pdbparse which can use just-parsed data to determine dynamically
    sized data following.
    """

    def __init__(
        self,
        context: interfaces.context.ContextInterface,
        location: str,
        database_name: Optional[str] = None,
        progress_callback: constants.ProgressCallback = None,
    ) -> None:
        self._layer_name, self._context = self.load_pdb_layer(context, location)
        self._dbiheader: Optional[interfaces.objects.ObjectInterface] = None
        if not progress_callback:

            def progress_callback(x, y):
                return None

        self._progress_callback = progress_callback
        self.types: List[
            Tuple[
                interfaces.objects.ObjectInterface,
                Optional[str],
                interfaces.objects.ObjectInterface,
            ]
        ] = []
        self.bases: Dict[str, Any] = {}
        self.user_types: Dict[str, Any] = {}
        self.enumerations: Dict[str, Any] = {}
        self.symbols: Dict[str, Any] = {}
        self._omap_mapping: List[Tuple[int, int]] = []
        self._sections: List[interfaces.objects.ObjectInterface] = []
        self.metadata = {"format": "6.1.0", "windows": {}}
        self._database_name = database_name

    @property
    def context(self):
        return self._context

    @property
    def pdb_layer_name(self):
        return self._layer_name

    @classmethod
    def load_pdb_layer(
        cls, context: interfaces.context.ContextInterface, location: str
    ) -> Tuple[str, interfaces.context.ContextInterface]:
        """Loads a PDB file into a layer within the context and returns the
        name of the new layer.

        Note: the context may be changed by this method
        """
        physical_layer_name = context.layers.free_layer_name("FileLayer")
        physical_config_path = interfaces.configuration.path_join(
            "pdbreader", physical_layer_name
        )

        # Create the file layer
        # This must be specific to get us started, setup the config and run
        new_context = context.clone()
        new_context.config[
            interfaces.configuration.path_join(physical_config_path, "location")
        ] = location

        physical_layer = physical.FileLayer(
            new_context, physical_config_path, physical_layer_name
        )
        new_context.add_layer(physical_layer)

        # Add on the MSF format layer
        msf_layer_name = context.layers.free_layer_name("MSFLayer")
        msf_config_path = interfaces.configuration.path_join(
            "pdbreader", msf_layer_name
        )
        new_context.config[
            interfaces.configuration.path_join(msf_config_path, "base_layer")
        ] = physical_layer_name
        msf_layer = msf.PdbMultiStreamFormat(
            new_context, msf_config_path, msf_layer_name
        )
        new_context.add_layer(msf_layer)

        msf_layer.read_streams()

        return msf_layer_name, new_context

    def reset(self):
        self.bases = {}
        self.user_types = {}
        self.enumerations = {}
        self.symbols = {}
        self._sections = []
        self._omap_mapping = []

    def read_necessary_streams(self):
        """Read streams to populate the various internal components for a PDB
        table."""
        if not self.metadata["windows"].get("pdb", None):
            self.read_pdb_info_stream()
        if not self.user_types:
            self.read_tpi_stream()
        if not self.symbols:
            self.read_symbol_stream()

    def read_tpi_stream(self) -> None:
        """Reads the TPI type steam."""
        self.types = []

        type_references = self._read_info_stream(2, "TPI", self.types)

        self.process_types(type_references)

    def read_ipi_stream(self):
        """"""
        if not self._dbiheader:
            self.read_dbi_stream()

        vollog.debug("Reading IPI layer")

        ipi_list = []

        try:
            type_references = self._read_info_stream(4, "IPI", ipi_list)
            for name in type_references.keys():
                # This doesn't break, because we want to use the last string/pdbname in the list
                if name.endswith(".pdb"):
                    self._database_name = name.split("\\")[-1]
        except ValueError:
            return None

    def _read_info_stream(self, stream_number, stream_name, info_list):
        vollog.debug(f"Reading {stream_name}")
        info_layer = self._context.layers.get(
            self._layer_name + "_stream" + str(stream_number), None
        )
        if not info_layer:
            raise ValueError(f"No {stream_name} stream available")
        module = self._context.module(
            module_name=info_layer.pdb_symbol_table,
            layer_name=info_layer.name,
            offset=0,
        )
        header = module.object(object_type="TPI_HEADER", offset=0)
        # Check the header
        if not (56 <= header.header_size < 1024):
            raise ValueError(f"{stream_name} Stream Header size outside normal bounds")
        if header.index_min < 4096:
            raise ValueError(
                f"Minimum {stream_name} index is 4096, found: {header.index_min}"
            )
        if header.index_max < header.index_min:
            raise ValueError(
                f"Maximum {stream_name} index is smaller than minimum TPI index, found: {header.index_max} < {header.index_min} "
            )
        # Reset the state
        info_references: Dict[str, int] = {}
        offset = header.header_size
        # Ensure we use the same type everywhere
        length_type = "unsigned short"
        length_len = module.get_type(length_type).size
        info_index = 1
        while info_layer.maximum_address - offset > 0:
            self._progress_callback(
                offset * 100 / info_layer.maximum_address, "Reading TPI layer"
            )
            length = module.object(object_type=length_type, offset=offset)
            if not isinstance(length, int):
                raise TypeError("Non-integer length provided")
            offset += length_len
            output, consumed = self.consume_type(module, offset, length)
            leaf_type, name, value = output
            for tag_type in ["unnamed", "anonymous"]:
                if name == f"<{tag_type}-tag>" or name == f"__{tag_type}":
                    name = f"__{tag_type}_" + hex(len(info_list) + 0x1000)[2:]
            if name:
                info_references[name] = len(info_list)
            info_list.append((leaf_type, name, value))
            offset += length
            info_index += 1
            # Since types can only refer to earlier types, assigning the name at this point is fine
        if info_layer.maximum_address - offset != 0:
            raise ValueError("Type values did not fill the TPI stream correctly")
        return info_references

    def read_dbi_stream(self) -> None:
        """Reads the DBI Stream."""
        vollog.debug("Reading DBI stream")
        dbi_layer = self._context.layers.get(self._layer_name + "_stream3", None)
        if not dbi_layer:
            raise ValueError("No DBI stream available")
        module = self._context.module(
            module_name=dbi_layer.pdb_symbol_table, layer_name=dbi_layer.name, offset=0
        )
        self._dbiheader = module.object(object_type="DBI_HEADER", offset=0)

        if not self._dbiheader:
            raise ValueError("DBI Header could not be read")

        # Skip past sections we don't care about to get to the DBG header
        dbg_hdr_offset = (
            self._dbiheader.vol.size
            + self._dbiheader.module_size
            + self._dbiheader.secconSize
            + self._dbiheader.secmapSize
            + self._dbiheader.filinfSize
            + self._dbiheader.tsmapSize
            + self._dbiheader.ecinfoSize
        )
        self._dbidbgheader = module.object(
            object_type="DBI_DBG_HEADER", offset=dbg_hdr_offset
        )

        self._sections = []
        self._omap_mapping = []

        if self._dbidbgheader.snSectionHdrOrig != -1:
            section_orig_layer_name = (
                self._layer_name + "_stream" + str(self._dbidbgheader.snSectionHdrOrig)
            )
            consumed, length = (
                0,
                self.context.layers[section_orig_layer_name].maximum_address,
            )
            while consumed < length:
                section = self.context.object(
                    dbi_layer.pdb_symbol_table
                    + constants.BANG
                    + "IMAGE_SECTION_HEADER",
                    offset=consumed,
                    layer_name=section_orig_layer_name,
                )
                self._sections.append(section)
                consumed += section.vol.size

            if self._dbidbgheader.snOmapFromSrc != -1:
                omap_layer_name = (
                    self._layer_name + "_stream" + str(self._dbidbgheader.snOmapFromSrc)
                )
                length = self.context.layers[omap_layer_name].maximum_address
                data = self.context.layers[omap_layer_name].read(0, length)
                # For speed we don't use the framework to read this (usually sizeable) data
                for i in range(0, length, 8):
                    self._omap_mapping.append(
                        (
                            int.from_bytes(data[i : i + 4], byteorder="little"),
                            int.from_bytes(data[i + 4 : i + 8], byteorder="little"),
                        )
                    )
        elif self._dbidbgheader.snSectionHdr != -1:
            section_layer_name = (
                self._layer_name + "_stream" + str(self._dbidbgheader.snSectionHdr)
            )
            consumed, length = (
                0,
                self.context.layers[section_layer_name].maximum_address,
            )
            while consumed < length:
                section = self.context.object(
                    dbi_layer.pdb_symbol_table
                    + constants.BANG
                    + "IMAGE_SECTION_HEADER",
                    offset=consumed,
                    layer_name=section_layer_name,
                )
                self._sections.append(section)
                consumed += section.vol.size

    def read_symbol_stream(self):
        """Reads in the symbol stream."""
        self.symbols = {}

        if not self._dbiheader:
            self.read_dbi_stream()

        vollog.debug("Reading Symbols")

        symrec_layer = self._context.layers.get(
            self._layer_name + "_stream" + str(self._dbiheader.symrecStream), None
        )
        if not symrec_layer:
            raise ValueError("No SymRec stream available")
        module = self._context.module(
            module_name=symrec_layer.pdb_symbol_table,
            layer_name=symrec_layer.name,
            offset=0,
        )

        offset = 0
        max_address = symrec_layer.maximum_address

        while offset < max_address:
            self._progress_callback(offset * 100 / max_address, "Reading Symbol layer")
            sym = module.object(object_type="GLOBAL_SYMBOL", offset=offset)
            leaf_type = module.object(
                object_type="unsigned short", offset=sym.leaf_type.vol.offset
            )
            name = None
            address = None
            if sym.segment < len(self._sections):
                if leaf_type == 0x1009:
                    # v2 symbol (pascal-string)
                    name = self.parse_string(
                        sym.name, True, sym.length - sym.vol.size + 2
                    )
                    address = (
                        self._sections[sym.segment - 1].VirtualAddress + sym.offset
                    )
                elif leaf_type == 0x110E or leaf_type == 0x1127:
                    # v3 symbol (c-string)
                    name = self.parse_string(
                        sym.name, False, sym.length - sym.vol.size + 2
                    )
                    address = (
                        self._sections[sym.segment - 1].VirtualAddress + sym.offset
                    )
                else:
                    vollog.debug(f"Only v2 and v3 symbols are supported: {leaf_type:x}")
            if name:
                if self._omap_mapping:
                    address = self.omap_lookup(address)
                stripped_name = self.name_strip(name)
                self.symbols[stripped_name] = {"address": address}
                if name != self.name_strip(name):
                    self.symbols[stripped_name]["linkage_name"] = name
            offset += sym.length + 2  # Add on length itself

    def read_pdb_info_stream(self):
        """Reads in the pdb information stream."""
        if not self._dbiheader:
            self.read_dbi_stream()
        if self._database_name is None:
            self.read_ipi_stream()

        vollog.debug("Reading PDB Info")
        pdb_info_layer = self._context.layers.get(self._layer_name + "_stream1", None)
        if not pdb_info_layer:
            raise ValueError("No PDB Info Stream available")
        module = self._context.module(
            module_name=pdb_info_layer.pdb_symbol_table,
            layer_name=pdb_info_layer.name,
            offset=0,
        )
        pdb_info = module.object(object_type="PDB_INFORMATION", offset=0)

        self.metadata["windows"]["pdb"] = {
            "GUID": self.convert_bytes_to_guid(pdb_info.GUID),
            "age": self._dbiheader.age,
            "database": self._database_name or "unknown.pdb",
            "machine_type": self._dbiheader.machine,
        }

    def convert_bytes_to_guid(self, original: bytes) -> str:
        """Convert the bytes to the correct ordering for a GUID."""
        orig_guid_list = [x for x in original]
        guid_list = []
        for i in [3, 2, 1, 0, 5, 4, 7, 6, 8, 9, 10, 11, 12, 13, 14, 15]:
            guid_list.append(orig_guid_list[i])
        return str(binascii.hexlify(bytes(guid_list)), "latin-1").upper()

    # SYMBOL HANDLING CODE

    def omap_lookup(self, address):
        """Looks up an address using the omap mapping."""
        pos = bisect(self._omap_mapping, (address, -1))
        if self._omap_mapping[pos][0] > address:
            pos -= 1

        if not self._omap_mapping[pos][1]:
            return 0
        return self._omap_mapping[pos][1] + (address - self._omap_mapping[pos][0])

    def name_strip(self, name):
        """Strips unnecessary components from the start of a symbol name."""
        new_name = name

        if new_name[:1] in ["_", "@", "\u007f"]:
            new_name = new_name[1:]

        name_array = new_name.split("@")
        if len(name_array) == 2:
            if name_array[1].isnumeric() and name_array[0][0] != "?":
                new_name = name_array[0]
            else:
                new_name = name

        return new_name

    def get_json(self):
        """Returns the intermediate format JSON data from this pdb file."""
        self.read_necessary_streams()

        # Set the time/datestamp for the output
        self.metadata["producer"] = {
            "datetime": datetime.datetime.now().isoformat(),
            "name": "volatility3",
            "version": constants.PACKAGE_VERSION,
        }

        return {
            "user_types": self.user_types,
            "enums": self.enumerations,
            "base_types": self.bases,
            "symbols": self.symbols,
            "metadata": self.metadata,
        }

    def get_type_from_index(self, index: int) -> Union[List[Any], Dict[str, Any]]:
        """Takes a type index and returns appropriate dictionary."""
        if index < 0x1000:
            base_name, base = primitives[index & 0xFF]
            self.bases[base_name] = base
            result: Union[List[Dict[str, Any]], Dict[str, Any]] = {
                "kind": "base",
                "name": base_name,
            }
            indirection = index & 0xF00
            if indirection:
                pointer_name, pointer_base = indirections[indirection]
                if (
                    self.bases.get("pointer", None)
                    and self.bases["pointer"] == pointer_base
                ):
                    result = {"kind": "pointer", "subtype": result}
                else:
                    self.bases[pointer_name] = pointer_base
                    result = {
                        "kind": "pointer",
                        "base": pointer_name,
                        "subtype": result,
                    }
            return result
        else:
            leaf_type, name, value = self.types[index - 0x1000]
            result = {"kind": "struct", "name": name}
            if leaf_type in [leaf_type.LF_MODIFIER]:
                result = self.get_type_from_index(value.subtype_index)
            elif leaf_type in [
                leaf_type.LF_ARRAY,
                leaf_type.LF_ARRAY_ST,
                leaf_type.LF_STRIDED_ARRAY,
            ]:
                result = {
                    "count": ForwardArrayCount(value.size, value.element_type),
                    "kind": "array",
                    "subtype": self.get_type_from_index(value.element_type),
                }
            elif leaf_type in [leaf_type.LF_BITFIELD]:
                result = {
                    "kind": "bitfield",
                    "type": self.get_type_from_index(value.underlying_type),
                    "bit_length": value.length,
                    "bit_position": value.position,
                }
            elif leaf_type in [leaf_type.LF_POINTER]:
                # Since we use the base['pointer'] to set the size for pointers, update it and check we don't get conflicts
                size = self.get_size_from_index(index)
                if self.bases.get("pointer", None) is None:
                    self.bases["pointer"] = {
                        "endian": "little",
                        "kind": "int",
                        "signed": False,
                        "size": size,
                    }
                else:
                    if size != self.bases["pointer"]["size"]:
                        raise ValueError("Native pointers with different sizes!")
                result = {
                    "kind": "pointer",
                    "subtype": self.get_type_from_index(value.subtype_index),
                }
            elif leaf_type in [leaf_type.LF_PROCEDURE]:
                return {"kind": "function"}
            elif leaf_type in [leaf_type.LF_UNION]:
                result = {"kind": "union", "name": name}
            elif leaf_type in [leaf_type.LF_ENUM]:
                result = {"kind": "enum", "name": name}
            elif leaf_type in [leaf_type.LF_FIELDLIST]:
                result = value
            elif not name:
                raise ValueError("No name for structure that should be named")
            return result

    def get_size_from_index(self, index: int) -> int:
        """Returns the size of the structure based on the type index
        provided."""
        result = -1
        name: Optional[str] = ""
        if index < 0x1000:
            if index & 0xF00:
                _, base = indirections[index & 0xF00]
            else:
                _, base = primitives[index & 0xFF]
            result = base["size"]
        else:
            leaf_type, name, value = self.types[index - 0x1000]
            if leaf_type in [
                leaf_type.LF_UNION,
                leaf_type.LF_CLASS,
                leaf_type.LF_CLASS_ST,
                leaf_type.LF_STRUCTURE,
                leaf_type.LF_STRUCTURE_ST,
                leaf_type.LF_INTERFACE,
                leaf_type.LF_CLASS_VS19,
                leaf_type.LF_STRUCTURE_VS19,
            ]:
                if not value.properties.forward_reference:
                    result = value.size
            elif leaf_type in [
                leaf_type.LF_ARRAY,
                leaf_type.LF_ARRAY_ST,
                leaf_type.LF_STRIDED_ARRAY,
            ]:
                result = value.size
            elif leaf_type in [
                leaf_type.LF_MODIFIER,
                leaf_type.LF_ENUM,
                leaf_type.LF_ARGLIST,
            ]:
                result = self.get_size_from_index(value.subtype_index)
            elif leaf_type in [leaf_type.LF_MEMBER]:
                result = self.get_size_from_index(value.field_type)
            elif leaf_type in [leaf_type.LF_BITFIELD]:
                result = self.get_size_from_index(value.underlying_type)
            elif leaf_type in [leaf_type.LF_POINTER]:
                result = value.size
                if not result:
                    if value.pointer_type == 0x0A:
                        return 4
                    elif value.pointer_type == 0x0C:
                        return 8
                    else:
                        raise ValueError("Pointer size could not be determined")
            elif leaf_type in [leaf_type.LF_PROCEDURE]:
                raise ValueError("LF_PROCEDURE size could not be identified")
            else:
                raise ValueError(
                    f"Unable to determine size of leaf_type {leaf_type.lookup()}"
                )
        if result <= 0:
            raise ValueError(f"Invalid size identified: {index} ({name})")
        return result

    ### TYPE HANDLING CODE

    def process_types(self, type_references: Dict[str, int]) -> None:
        """Reads the TPI and symbol streams to populate the reader's
        variables."""

        self.bases = {}
        self.user_types = {}
        self.enumerations = {}

        max_len = len(self.types)
        for index in range(max_len):
            self._progress_callback(index * 100 / max_len, "Processing types")
            leaf_type, name, value = self.types[index]
            if leaf_type in [
                leaf_type.LF_CLASS,
                leaf_type.LF_CLASS_ST,
                leaf_type.LF_STRUCTURE,
                leaf_type.LF_STRUCTURE_ST,
                leaf_type.LF_INTERFACE,
                leaf_type.LF_CLASS_VS19,
                leaf_type.LF_STRUCTURE_VS19,
            ]:
                if not value.properties.forward_reference and name:
                    self.user_types[name] = {
                        "kind": "struct",
                        "size": value.size,
                        "fields": self.convert_fields(value.fields - 0x1000),
                    }
            elif leaf_type in [leaf_type.LF_UNION]:
                if not value.properties.forward_reference and name:
                    # Deal with UNION types
                    self.user_types[name] = {
                        "kind": "union",
                        "size": value.size,
                        "fields": self.convert_fields(value.fields - 0x1000),
                    }
            elif leaf_type in [leaf_type.LF_ENUM]:
                if not value.properties.forward_reference and name:
                    base = self.get_type_from_index(value.subtype_index)
                    if not isinstance(base, Dict):
                        raise ValueError("Invalid base type returned for Enumeration")
                    constants = self.get_type_from_index(value.fields)
                    if not isinstance(constants, list):
                        raise ValueError("Enumeration fields type not a list")
                    self.enumerations[name] = {
                        "base": base["name"],
                        "size": self.get_size_from_index(value.subtype_index),
                        "constants": dict(
                            [(name, enum.value) for _, name, enum in constants]
                        ),
                    }

        # Re-run through for ForwardSizeReferences
        self.user_types = self.replace_forward_references(
            self.user_types, type_references
        )

    type_handlers = {
        # Leaf_type: ('Structure', has_name, value_attribute)
        "LF_CLASS": ("LF_STRUCTURE", True, "size"),
        "LF_CLASS_ST": ("LF_STRUCTURE", True, "size"),
        "LF_STRUCTURE": ("LF_STRUCTURE", True, "size"),
        "LF_STRUCTURE_ST": ("LF_STRUCTURE", True, "size"),
        "LF_INTERFACE": ("LF_STRUCTURE", True, "size"),
        "LF_CLASS_VS19": ("LF_STRUCTURE_VS19", True, "size"),
        "LF_STRUCTURE_VS19": ("LF_STRUCTURE_VS19", True, "size"),
        "LF_MEMBER": ("LF_MEMBER", True, "offset"),
        "LF_MEMBER_ST": ("LF_MEMBER", True, "offset"),
        "LF_ARRAY": ("LF_ARRAY", True, "size"),
        "LF_ARRAY_ST": ("LF_ARRAY", True, "size"),
        "LF_STRIDED_ARRAY": ("LF_ARRAY", True, "size"),
        "LF_ENUMERATE": ("LF_ENUMERATE", True, "value"),
        "LF_ARGLIST": ("LF_ENUM", True, None),
        "LF_ENUM": ("LF_ENUM", True, None),
        "LF_UNION": ("LF_UNION", True, None),
        "LF_STRING_ID": ("LF_STRING_ID", True, None),
        "LF_FUNC_ID": ("LF_FUNC_ID", True, None),
        "LF_MODIFIER": ("LF_MODIFIER", False, None),
        "LF_POINTER": ("LF_POINTER", False, None),
        "LF_PROCEDURE": ("LF_PROCEDURE", False, None),
        "LF_FIELDLIST": ("LF_FIELDLIST", False, None),
        "LF_BITFIELD": ("LF_BITFIELD", False, None),
        "LF_UDT_SRC_LINE": ("LF_UDT_SRC_LINE", False, None),
        "LF_UDT_MOD_SRC_LINE": ("LF_UDT_MOD_SRC_LINE", False, None),
        "LF_BUILDINFO": ("LF_BUILDINFO", False, None),
    }

    def consume_type(
        self, module: interfaces.context.ModuleInterface, offset: int, length: int
    ) -> Tuple[
        Tuple[
            Optional[interfaces.objects.ObjectInterface],
            Optional[str],
            Union[None, List, interfaces.objects.ObjectInterface],
        ],
        int,
    ]:
        """Returns a (leaf_type, name, object) Tuple for a type, and the number
        of bytes consumed."""
        leaf_type = self.context.object(
            module.get_enumeration("LEAF_TYPE"),
            layer_name=module._layer_name,
            offset=offset,
        )
        consumed = leaf_type.vol.base_type.size
        remaining = length - consumed

        type_handler, has_name, value_attribute = self.type_handlers.get(
            leaf_type.lookup(), ("LF_UNKNOWN", False, None)
        )

        if type_handler in ["LF_FIELDLIST"]:
            sub_length = remaining
            sub_offset = offset + consumed
            fields = []
            while length > consumed:
                subfield, sub_consumed = self.consume_type(
                    module, sub_offset, sub_length
                )
                sub_consumed += self.consume_padding(
                    module.layer_name, sub_offset + sub_consumed
                )
                sub_length -= sub_consumed
                sub_offset += sub_consumed
                consumed += sub_consumed
                fields.append(subfield)
            result = leaf_type, None, fields
        elif type_handler in ["LF_BUILDINFO"]:
            parsed_obj = module.object(
                object_type=type_handler, offset=offset + consumed
            )
            parsed_obj.arguments.count = parsed_obj.count
            consumed += parsed_obj.arguments.vol.size
            result = leaf_type, None, parsed_obj
        elif type_handler in self.type_handlers:
            parsed_obj = module.object(
                object_type=type_handler, offset=offset + consumed
            )
            current_consumed = remaining
            if has_name:
                name_offset = parsed_obj.name.vol.offset - parsed_obj.vol.offset
                if value_attribute:
                    name, value, excess = self.determine_extended_value(
                        leaf_type,
                        getattr(parsed_obj, value_attribute),
                        module,
                        remaining - name_offset,
                    )
                    setattr(parsed_obj, value_attribute, value)
                    current_consumed = parsed_obj.vol.size + len(name) + 1 + excess
                else:
                    name = self.parse_string(
                        parsed_obj.name,
                        leaf_type < leaf_type.LF_ST_MAX,
                        size=remaining - name_offset,
                    )
                parsed_obj.name = name
            else:
                name = None
            result = leaf_type, name, parsed_obj
            consumed += current_consumed
        else:
            raise TypeError(f"Unhandled leaf_type: {leaf_type}")

        return result, consumed

    def consume_padding(self, layer_name: str, offset: int) -> int:
        """Returns the amount of padding used between fields."""
        val = self.context.layers[layer_name].read(offset, 1)
        if not ((val[0] & 0xF0) == 0xF0):
            return 0
        return int(val[0]) & 0x0F

    def convert_fields(self, fields: int) -> Dict[Optional[str], Dict[str, Any]]:
        """Converts a field list into a list of fields."""
        result: Dict[Optional[str], Dict[str, Any]] = {}
        _, _, fields_struct = self.types[fields]
        if not isinstance(fields_struct, list):
            vollog.warning("Fields structure did not contain a list of fields")
            return result
        for field in fields_struct:
            _, name, member = field
            result[name] = {
                "offset": member.offset,
                "type": self.get_type_from_index(member.field_type),
            }
        return result

    def replace_forward_references(self, types, type_references):
        """Finds all ForwardArrayCounts and calculates them once
        ForwardReferences have been resolved."""
        if isinstance(types, dict):
            for k, v in types.items():
                types[k] = self.replace_forward_references(v, type_references)
        elif isinstance(types, list):
            new_types = []
            for v in types:
                new_types.append(self.replace_forward_references(v, type_references))
            types = new_types
        elif isinstance(types, ForwardArrayCount):
            element_type = types.element_type
            # If we're a forward array count, we need to do the calculation now after all the types have been processed
            loop = True
            while loop:
                loop = False
                if element_type > 0x1000:
                    _, name, toplevel_type = self.types[element_type - 0x1000]
                    # If there's no name, the original size is probably fine as long as we're not indirect (LF_MODIFIER)
                    if (
                        not name
                        and isinstance(
                            toplevel_type, interfaces.objects.ObjectInterface
                        )
                        and toplevel_type.vol.type_name.endswith("LF_MODIFIER")
                    ):
                        # We have check they don't point to a forward reference, so we go round again with the subtype
                        element_type = toplevel_type.subtype_index
                        loop = True
                    elif name:
                        # If there is a name, look it up so we're not using a reference but the real thing
                        element_type = type_references[name] + 0x1000
            return types.size // self.get_size_from_index(element_type)
        return types

    # COMMON CODE

    @staticmethod
    def parse_string(
        structure: interfaces.objects.ObjectInterface,
        parse_as_pascal: bool = False,
        size: int = 0,
    ) -> str:
        """Consumes either a c-string or a pascal string depending on the
        leaf_type."""
        if not parse_as_pascal:
            name = structure.cast("string", max_length=size, encoding="latin-1")
        else:
            name = structure.cast("pascal_string")
            name = name.string.cast(
                "string", max_length=name.length, encoding="latin-1"
            )
        return str(name)

    def determine_extended_value(
        self,
        leaf_type: interfaces.objects.ObjectInterface,
        value: interfaces.objects.ObjectInterface,
        module: interfaces.context.ModuleInterface,
        length: int,
    ) -> Tuple[str, interfaces.objects.ObjectInterface, int]:
        """Reads a value and potentially consumes more data to construct the
        value."""
        excess = 0
        if value >= leaf_type.LF_CHAR:
            sub_leaf_type = self.context.object(
                self.context.symbol_space.get_enumeration(leaf_type.vol.type_name),
                layer_name=leaf_type.vol.layer_name,
                offset=value.vol.offset,
            )
            # Set the offset at just after the previous size type
            offset = value.vol.offset + value.vol.data_format.length
            if sub_leaf_type in [leaf_type.LF_CHAR]:
                value = module.object(object_type="char", offset=offset)
            elif sub_leaf_type in [leaf_type.LF_SHORT]:
                value = module.object(object_type="short", offset=offset)
            elif sub_leaf_type in [leaf_type.LF_USHORT]:
                value = module.object(object_type="unsigned short", offset=offset)
            elif sub_leaf_type in [leaf_type.LF_LONG]:
                value = module.object(object_type="long", offset=offset)
            elif sub_leaf_type in [leaf_type.LF_ULONG]:
                value = module.object(object_type="unsigned long", offset=offset)
            else:
                raise TypeError("Unexpected extended value type")
            excess = value.vol.data_format.length
            # Updated the consume/offset counters
        name = module.object(
            object_type="string", offset=value.vol.offset + value.vol.data_format.length
        )
        name_str = self.parse_string(
            name, leaf_type < leaf_type.LF_ST_MAX, size=length - excess
        )
        return name_str, value, excess


class PdbRetreiver:
    def retreive_pdb(
        self,
        guid: str,
        file_name: str,
        progress_callback: constants.ProgressCallback = None,
    ) -> Optional[str]:
        vollog.info("Download PDB file...")
        file_name = ".".join(file_name.split(".")[:-1] + ["pdb"])
        for sym_url in [constants.SYMBOL_SERVER_URL]:
            url = sym_url + f"/{file_name}/{guid}/"

            result = None
            for suffix in [file_name, file_name[:-1] + "_"]:
                try:
                    vollog.debug(f"Attempting to retrieve {url + suffix}")
                    # We have to cache this because the file is opened by a layer and we can't control whether that caches
                    with resources.ResourceAccessor(progress_callback).open(
                        url + suffix
                    ) as fp:
                        fp.read(10)
                        result = True
                except (error.HTTPError, error.URLError) as excp:
                    vollog.debug(f"Failed with {excp}")
                if result:
                    break
        if progress_callback is not None:
            progress_callback(100, f"Downloading {url + suffix}")
        if not result:
            return None
        return url + suffix


if __name__ == "__main__":
    import argparse

    class PrintedProgress:
        """A progress handler that prints the progress value and the
        description onto the command line."""

        def __init__(self):
            self._max_message_len = 0

        def __call__(
            self, progress: Union[int, float], description: Optional[str] = None
        ):
            """A simple function for providing text-based feedback.

            .. warning:: Only for development use.

            Args:
                progress: Percentage of progress of the current procedure
            """
            message = f"\rProgress: {round(progress, 2): 7.2f}\t\t{description or ''}"
            message_len = len(message)
            self._max_message_len = max([self._max_message_len, message_len])
            print(message, end=(" " * (self._max_message_len - message_len)) + "\r")

    parser = argparse.ArgumentParser(
        description="Read PDB files and convert to Volatility 3 Intermediate Symbol Format"
    )
    parser.add_argument(
        "-o",
        "--output",
        metavar="OUTPUT",
        help="Filename for data output",
        default=None,
    )
    file_group = parser.add_argument_group(
        "file", description="File-based conversion of PDB to ISF"
    )
    file_group.add_argument(
        "-f", "--file", metavar="FILE", help="PDB file to translate to ISF"
    )
    data_group = parser.add_argument_group(
        "data", description="Convert based on a GUID and filename pattern"
    )
    data_group.add_argument(
        "-p",
        "--pattern",
        metavar="PATTERN",
        help="Filename pattern to recover PDB file",
    )
    data_group.add_argument(
        "-g",
        "--guid",
        metavar="GUID",
        help="GUID + Age string for the required PDB file",
        default=None,
    )
    data_group.add_argument(
        "-k",
        "--keep",
        action="store_true",
        default=False,
        help="Keep the downloaded PDB file",
    )
    args = parser.parse_args()

    pg_cb = PrintedProgress()

    delfile = False
    filename = None
    if args.guid is not None and args.pattern is not None:
        filename = PdbRetreiver().retreive_pdb(
            guid=args.guid, file_name=args.pattern, progress_callback=pg_cb
        )
        if filename is None:
            parser.error("PDB file could not be retrieved from the internet")
        if parse.urlparse(filename, "file").scheme == "file":
            delfile = True
    elif args.file:
        filename = args.file
    else:
        parser.error("No GUID/pattern or file provided")

    if not filename:
        parser.error("No suitable filename provided or retrieved")

    ctx = contexts.Context()
    url = parse.urlparse(filename, scheme="file")
    if url.scheme == "file":
        if not os.path.exists(filename):
            parser.error(f"File {filename} does not exists")
        location = "file:" + request.pathname2url(os.path.abspath(filename))
    else:
        location = filename

    convertor = PdbReader(
        ctx, location, database_name=args.pattern, progress_callback=pg_cb
    )

    converted_json = convertor.get_json()
    if args.output is None:
        if args.guid:
            guid = args.guid[:-1]
            age = args.guid[-1:]
        else:
            guid = converted_json["metadata"]["windows"]["pdb"]["GUID"]
            age = converted_json["metadata"]["windows"]["pdb"]["age"]
        args.output = f"{guid}-{age}.json.xz"

    output_url = os.path.abspath(args.output)

    open_method = open
    if args.output.endswith(".gz"):
        open_method = gzip.open
    elif args.output.endswith(".bz2"):
        open_method = bz2.open
    elif args.output.endswith(".xz"):
        open_method = lzma.open

    with open_method(output_url, "wb") as f:
        json_string = json.dumps(converted_json, indent=2, sort_keys=True)
        f.write(bytes(json_string, "latin-1"))

    if args.keep:
        print(f"Temporary PDB file: {filename}")
    elif delfile:
        os.remove(filename)



# This file is Copyright 2024 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging
import datetime
from typing import List, Generator, Tuple

from volatility3.framework import constants
from volatility3.framework import interfaces, symbols, exceptions
from volatility3.framework import renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.interfaces import configuration
from volatility3.framework.renderers import format_hints, conversion
from volatility3.framework.symbols import intermed
from volatility3.plugins import timeliner
from volatility3.plugins.windows import modules

vollog = logging.getLogger(__name__)


class UnloadedModules(interfaces.plugins.PluginInterface, timeliner.TimeLinerInterface):
    """Lists the unloaded kernel modules."""

    _required_framework_version = (2, 0, 0)
    _version = (2, 0, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="timeliner",
                component=timeliner.TimeLinerInterface,
                version=(1, 0, 0),
            ),
            requirements.VersionRequirement(
                name="modules", component=modules.Modules, version=(3, 0, 0)
            ),
        ]

    @classmethod
    def create_unloadedmodules_table(
        cls,
        context: interfaces.context.ContextInterface,
        symbol_table: str,
        config_path: str,
    ) -> str:
        """Creates a symbol table for the unloaded modules.

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            symbol_table: The name of an existing symbol table containing the kernel symbols
            config_path: The configuration path within the context of the symbol table to create

        Returns:
            The name of the constructed unloaded modules table
        """
        native_types = context.symbol_space[symbol_table].natives
        is_64bit = symbols.symbol_table_is_64bit(
            context=context, symbol_table_name=symbol_table
        )
        table_mapping = {"nt_symbols": symbol_table}

        if is_64bit:
            symbol_filename = "unloadedmodules-x64"
        else:
            symbol_filename = "unloadedmodules-x86"

        return intermed.IntermediateSymbolTable.create(
            context,
            configuration.path_join(config_path, "unloadedmodules"),
            "windows",
            symbol_filename,
            native_types=native_types,
            table_mapping=table_mapping,
        )

    @classmethod
    def list_unloadedmodules(
        cls,
        context: interfaces.context.ContextInterface,
        kernel_module_name: str,
        unloadedmodule_table_name: str,
    ) -> Generator[Tuple[str, int, int, datetime.datetime], None, None]:
        """Lists all the unloaded modules in the primary layer.

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            layer_name: The name of the layer on which to operate
            symbol_table: The name of the table containing the kernel symbols

        Returns:
            A list of Unloaded Modules as retrieved from MmUnloadedDrivers
        """

        ntkrnlmp = context.modules[kernel_module_name]

        unloadedmodules_offset = ntkrnlmp.get_symbol("MmUnloadedDrivers").address
        unloadedmodules = ntkrnlmp.object(
            object_type="pointer",
            offset=unloadedmodules_offset,
            subtype="array",
        )
        is_64bit = symbols.symbol_table_is_64bit(
            context=context, symbol_table_name=ntkrnlmp.symbol_table_name
        )

        if is_64bit:
            unloaded_count_type = "unsigned long long"
        else:
            unloaded_count_type = "unsigned long"

        last_unloadedmodule_offset = ntkrnlmp.get_symbol("MmLastUnloadedDriver").address
        unloaded_count = ntkrnlmp.object(
            object_type=unloaded_count_type, offset=last_unloadedmodule_offset
        )

        # Bring down to default when smear present. Some samples had this completely broken
        if unloaded_count > 1024:
            vollog.warning(
                f"Smeared array count found {unloaded_count}. Defaulting to 1024 elements."
            )
            unloaded_count = 1024

        unloadedmodules_array = context.object(
            object_type=unloadedmodule_table_name
            + constants.BANG
            + "_UNLOADED_DRIVERS",
            layer_name=ntkrnlmp.layer_name,
            offset=unloadedmodules,
        )
        unloadedmodules_array.UnloadedDrivers.count = unloaded_count

        kernel_space_start = modules.Modules.get_kernel_space_start(
            context, kernel_module_name
        )

        address_mask = context.layers[ntkrnlmp.layer_name].address_mask

        for driver in unloadedmodules_array.UnloadedDrivers:
            # Mass testing led to dozens of samples backtracing on this plugin when
            # accessing members of modules coming out this list
            # Given how often temporary drivers load and unload on Win10+, I
            # assume the chance for smear is very high
            try:
                start_address = driver.StartAddress & address_mask
                end_address = driver.EndAddress & address_mask
                current_time = driver.CurrentTime
                driver_name = driver.Name.String
            except exceptions.InvalidAddressException:
                continue

            if (
                current_time > 1024
                and start_address > kernel_space_start
                and start_address & 0xFFF == 0x0
                and end_address & 0xFFF == 0x0
                and end_address > kernel_space_start
            ):
                yield driver_name, start_address, end_address, current_time

    def _generator(self):
        kernel = self.context.modules[self.config["kernel"]]

        if not kernel.has_symbol("MmUnloadedDrivers"):
            vollog.error(
                "The symbol table for this sample is missing the `MmUnloadedDrivers` symbol. Cannot proceed."
            )
            return

        if not kernel.has_symbol("MmLastUnloadedDriver"):
            vollog.error(
                "The symbol table for this sample is missing the `MmLastUnloadededDriver` symbol. Cannot proceed."
            )
            return

        unloadedmodule_table_name = self.create_unloadedmodules_table(
            self.context, kernel.symbol_table_name, self.config_path
        )

        for (
            driver_name,
            start_address,
            end_address,
            current_time,
        ) in self.list_unloadedmodules(
            self.context,
            self.config["kernel"],
            unloadedmodule_table_name,
        ):
            yield (
                0,
                (
                    driver_name,
                    format_hints.Hex(start_address),
                    format_hints.Hex(end_address),
                    conversion.wintime_to_datetime(current_time),
                ),
            )

    def generate_timeline(self):
        for row in self._generator():
            _depth, row_data = row
            description = f"Unloaded Module: {row_data[0]}"
            yield (description, timeliner.TimeLinerType.CHANGED, row_data[3])

    def run(self):
        return renderers.TreeGrid(
            [
                ("Name", str),
                ("StartAddress", format_hints.Hex),
                ("EndAddress", format_hints.Hex),
                ("Time", datetime.datetime),
            ],
            self._generator(),
        )



# This file is Copyright 2020 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import logging
from typing import List

from volatility3.framework import exceptions, interfaces
from volatility3.framework import renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.interfaces import plugins
from volatility3.framework.objects import utility
from volatility3.framework.renderers import format_hints
from volatility3.framework.symbols import mac
from volatility3.plugins.mac import lsmod

vollog = logging.getLogger(__name__)


class Socket_filters(plugins.PluginInterface):
    """Enumerates kernel socket filters."""

    _required_framework_version = (2, 0, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Kernel module for the OS",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="macutils", component=mac.MacUtilities, version=(1, 0, 0)
            ),
            requirements.VersionRequirement(
                name="lsmod", component=lsmod.Lsmod, version=(2, 0, 0)
            ),
        ]

    def _generator(self):
        kernel = self.context.modules[self.config["kernel"]]

        mods = lsmod.Lsmod.list_modules(self.context, self.config["kernel"])

        handlers = mac.MacUtilities.generate_kernel_handler_info(
            self.context, kernel.layer_name, kernel, mods
        )

        members_to_check = [
            "sf_unregistered",
            "sf_attach",
            "sf_detach",
            "sf_notify",
            "sf_getpeername",
            "sf_getsockname",
            "sf_data_in",
            "sf_data_out",
            "sf_connect_in",
            "sf_connect_out",
            "sf_bind",
            "sf_setoption",
            "sf_getoption",
            "sf_listen",
            "sf_ioctl",
        ]

        filter_list = kernel.object_from_symbol(symbol_name="sock_filter_head")

        for filter_container in mac.MacUtilities.walk_tailq(
            filter_list, "sf_global_next"
        ):
            current_filter = filter_container.sf_filter

            filter_name = utility.pointer_to_string(current_filter.sf_name, count=128)

            try:
                filter_socket = filter_container.sf_entry_head.sfe_socket.vol.offset
            except exceptions.InvalidAddressException:
                filter_socket = 0

            for member in members_to_check:
                check_addr = current_filter.member(attr=member)
                if check_addr == 0:
                    continue

                module_name, symbol_name = mac.MacUtilities.lookup_module_address(
                    self.context, handlers, check_addr
                )

                yield (
                    0,
                    (
                        format_hints.Hex(current_filter.vol.offset),
                        filter_name,
                        member,
                        format_hints.Hex(filter_socket),
                        format_hints.Hex(check_addr),
                        module_name,
                        symbol_name,
                    ),
                )

    def run(self):
        return renderers.TreeGrid(
            [
                ("Filter", format_hints.Hex),
                ("Name", str),
                ("Member", str),
                ("Socket", format_hints.Hex),
                ("Handler", format_hints.Hex),
                ("Module", str),
                ("Symbol", str),
            ],
            self._generator(),
        )



# This file is Copyright 2022 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

import logging

from typing import Iterator, List, Tuple

from volatility3.framework import exceptions, interfaces, renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.objects import utility
from volatility3.framework.renderers import format_hints
from volatility3.plugins.windows import pslist

vollog = logging.getLogger(__name__)


class JobLinks(interfaces.plugins.PluginInterface):
    """Print process job link information"""

    _required_framework_version = (2, 0, 0)
    _version = (1, 0, 0)

    @classmethod
    def get_requirements(cls) -> List[interfaces.configuration.RequirementInterface]:
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.BooleanRequirement(
                name="physical",
                description="Display physical offset instead of virtual",
                default=False,
                optional=True,
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(3, 0, 0)
            ),
        ]

    def _generator(self) -> Iterator[Tuple]:
        kernel = self.context.modules[self.config["kernel"]]

        memory = self.context.layers[kernel.layer_name]

        for proc in pslist.PsList.list_processes(
            context=self.context,
            kernel_module_name=self.config["kernel"],
        ):
            try:
                if not self.config["physical"]:
                    offset = proc.vol.offset
                else:
                    (_, _, offset, _, _) = list(
                        memory.mapping(offset=proc.vol.offset, length=0)
                    )[0]

                job = proc.Job.dereference()

                yield (
                    0,
                    (
                        format_hints.Hex(offset),
                        utility.array_to_string(proc.ImageFileName),
                        proc.UniqueProcessId,
                        proc.InheritedFromUniqueProcessId,
                        proc.get_session_id(),
                        job.SessionId,
                        proc.get_is_wow64(),
                        job.TotalProcesses,
                        job.ActiveProcesses,
                        job.TotalTerminatedProcesses,
                        renderers.NotApplicableValue(),
                        "(Original Process)",
                    ),
                )

                for entry in job.ProcessListHead.to_list(
                    proc.vol.type_name, "JobLinks"
                ):
                    if not self.config["physical"]:
                        offset = entry.vol.offset
                    else:
                        (_, _, offset, _, _) = list(
                            memory.mapping(offset=entry.vol.offset, length=0)
                        )[0]

                    yield (
                        1,
                        (
                            format_hints.Hex(offset),
                            utility.array_to_string(entry.ImageFileName),
                            entry.UniqueProcessId,
                            entry.InheritedFromUniqueProcessId,
                            entry.get_session_id(),
                            0,
                            entry.get_is_wow64(),
                            0,
                            0,
                            0,
                            "Yes",
                            entry.get_peb().ProcessParameters.ImagePathName.get_string(),
                        ),
                    )

            except exceptions.InvalidAddressException:
                continue

    def run(self) -> renderers.TreeGrid:
        offsettype = (
            "(V)"
            if not self.config.get("physical", pslist.PsList.PHYSICAL_DEFAULT)
            else "(P)"
        )

        return renderers.TreeGrid(
            [
                (f"Offset{offsettype}", format_hints.Hex),
                ("Name", str),
                ("PID", int),
                ("PPID", int),
                ("Sess", int),
                ("JobSess", int),
                ("Wow64", bool),
                ("Total", int),
                ("Active", int),
                ("Term", int),
                ("JobLink", str),
                ("Process", str),
            ],
            self._generator(),
        )



# This file is Copyright 2024 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#

# This module attempts to locate windows console histories.

import logging
import struct
from typing import Tuple, Generator, Set, Dict, Any, Optional

from volatility3.framework import interfaces
from volatility3.framework import renderers
from volatility3.framework.configuration import requirements
from volatility3.framework.layers import scanners
from volatility3.framework.objects import utility
from volatility3.framework.renderers import format_hints
from volatility3.plugins.windows import pslist, consoles


vollog = logging.getLogger(__name__)


class CmdScan(interfaces.plugins.PluginInterface):
    """Looks for Windows Command History lists"""

    _required_framework_version = (2, 4, 0)
    _version = (2, 0, 0)

    @classmethod
    def get_requirements(cls):
        # Since we're calling the plugin, make sure we have the plugin's requirements
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="pslist", component=pslist.PsList, version=(3, 0, 0)
            ),
            requirements.VersionRequirement(
                name="consoles", component=consoles.Consoles, version=(3, 0, 0)
            ),
            requirements.VersionRequirement(
                name="bytes_scanner",
                component=scanners.BytesScanner,
                version=(1, 0, 0),
            ),
            requirements.BooleanRequirement(
                name="no_registry",
                description="Don't search the registry for possible values of CommandHistorySize",
                optional=True,
                default=False,
            ),
            requirements.ListRequirement(
                name="max_history",
                element_type=int,
                description="CommandHistorySize values to search for.",
                optional=True,
                default=[50],
            ),
        ]

    @classmethod
    def get_filtered_vads(
        cls,
        conhost_proc: interfaces.context.ContextInterface,
        size_filter: Optional[int] = 0x40000000,
    ) -> Generator[Tuple[int, int], None, None]:
        """
        Returns vads of a process with size smaller than size_filter

        Args:
            conhost_proc: the process object for conhost.exe
            size_filter: size above which vads will not be returned

        Returns:
            A list of tuples of:
            vad_base: the base address
            vad_size: the size of the VAD
        """
        for vad in conhost_proc.get_vad_root().traverse():
            base = vad.get_start()
            if vad.get_size() < size_filter:
                yield (base, vad.get_size())

    @classmethod
    def get_command_history(
        cls,
        context: interfaces.context.ContextInterface,
        config_path: str,
        kernel_module_name: str,
        procs: Generator[interfaces.objects.ObjectInterface, None, None],
        max_history: Set[int],
    ) -> Tuple[
        interfaces.context.ContextInterface,
        interfaces.context.ContextInterface,
        Dict[str, Any],
    ]:
        """Gets the list of commands from each Command History structure

        Args:
            context: The context to retrieve required elements (layers, symbol tables) from
            config_path: The config path where to find symbol files
            procs: List of process objects
            max_history: An initial set of CommandHistorySize values

        Returns:
            The conhost process object, the command history structure, a dictionary of properties for
            that command history structure.
        """

        conhost_symbol_table = None

        for conhost_proc, proc_layer_name in consoles.Consoles.find_conhost_proc(procs):
            if not conhost_proc:
                vollog.info(
                    "Unable to find a valid conhost.exe process in the process list. Analysis cannot proceed."
                )
                continue
            vollog.debug(
                f"Found conhost process {conhost_proc} with pid {conhost_proc.UniqueProcessId}"
            )

            conhostexe_base, conhostexe_size = consoles.Consoles.find_conhostexe(
                conhost_proc
            )
            if not conhostexe_base:
                vollog.info(
                    "Unable to find the location of conhost.exe. Analysis cannot proceed."
                )
                continue
            vollog.debug(f"Found conhost.exe base at {conhostexe_base:#x}")

            proc_layer = context.layers[proc_layer_name]

            if conhost_symbol_table is None:
                conhost_symbol_table = consoles.Consoles.create_conhost_symbol_table(
                    context,
                    config_path,
                    kernel_module_name,
                    proc_layer_name,
                    conhostexe_base,
                )

            conhost_module = context.module(
                conhost_symbol_table, proc_layer_name, offset=conhostexe_base
            )
            command_count_max_offset = conhost_module.get_type(
                "_COMMAND_HISTORY"
            ).relative_child_offset("CommandCountMax")

            sections = cls.get_filtered_vads(conhost_proc)
            found_history_for_proc = False
            # scan for potential _COMMAND_HISTORY structures by using the CommandHistorySize
            for max_history_value in max_history:
                max_history_bytes = struct.pack("H", max_history_value)
                vollog.debug(
                    f"Scanning for CommandHistorySize value: {max_history_bytes}"
                )
                for address in proc_layer.scan(
                    context,
                    scanners.BytesScanner(max_history_bytes),
                    sections=sections,
                ):
                    command_history = None
                    command_history_properties = []

                    try:
                        command_history = conhost_module.object(
                            "_COMMAND_HISTORY",
                            offset=address - command_count_max_offset,
                            absolute=True,
                        )

                        if not command_history.is_valid(max_history_value):
                            continue

                        vollog.debug(
                            f"Getting Command History properties for {command_history}"
                        )
                        command_history_properties.append(
                            {
                                "level": 0,
                                "name": "_COMMAND_HISTORY",
                                "address": command_history.vol.offset,
                                "data": None,
                            }
                        )
                        command_history_properties.append(
                            {
                                "level": 1,
                                "name": "_COMMAND_HISTORY.Application",
                                "address": command_history.Application.vol.offset,
                                "data": command_history.get_application(),
                            }
                        )
                        command_history_properties.append(
                            {
                                "level": 1,
                                "name": "_COMMAND_HISTORY.ProcessHandle",
                                "address": command_history.ConsoleProcessHandle.ProcessHandle.vol.offset,
                                "data": hex(
                                    command_history.ConsoleProcessHandle.ProcessHandle
                                ),
                            }
                        )
                        command_history_properties.append(
                            {
                                "level": 1,
                                "name": "_COMMAND_HISTORY.CommandCount",
                                "address": None,
                                "data": command_history.CommandCount,
                            }
                        )
                        command_history_properties.append(
                            {
                                "level": 1,
                                "name": "_COMMAND_HISTORY.LastDisplayed",
                                "address": command_history.LastDisplayed.vol.offset,
                                "data": command_history.LastDisplayed,
                            }
                        )
                        command_history_properties.append(
                            {
                                "level": 1,
                                "name": "_COMMAND_HISTORY.CommandCountMax",
                                "address": command_history.CommandCountMax.vol.offset,
                                "data": command_history.CommandCountMax,
                            }
                        )
                        command_history_properties.append(
                            {
                                "level": 1,
                                "name": "_COMMAND_HISTORY.CommandBucket",
                                "address": command_history.CommandBucket.vol.offset,
                                "data": "",
                            }
                        )

                        for (
                            cmd_index,
                            bucket_cmd,
                        ) in command_history.scan_command_bucket():
                            try:
                                command_history_properties.append(
                                    {
                                        "level": 2,
                                        "name": f"_COMMAND_HISTORY.CommandBucket_Command_{cmd_index}",
                                        "address": bucket_cmd.vol.offset,
                                        "data": bucket_cmd.get_command_string(),
                                    }
                                )
                            except Exception as e:
                                vollog.debug(
                                    f"reading {bucket_cmd} encountered exception {e}"
                                )
                    except Exception as e:
                        vollog.debug(
                            f"reading {command_history} encountered exception {e}"
                        )

                    if command_history and command_history_properties:
                        found_history_for_proc = True
                        yield conhost_proc, command_history, command_history_properties

            # if found_history_for_proc is still False, then none of the scanned locations found
            # a valid _COMMAND_HISTORY for the process, so yield the process and some empty data
            # so the process can at least be reported that it was found with no history
            if not found_history_for_proc:
                yield conhost_proc, command_history or None, []

    def _generator(
        self, procs: Generator[interfaces.objects.ObjectInterface, None, None]
    ):
        """
        Generates the command history to use in rendering

        Args:
            procs: the process list filtered to conhost.exe instances
        """

        max_history = set(self.config.get("max_history", [50]))
        no_registry = self.config.get("no_registry")

        if no_registry is False:
            max_history, _ = consoles.Consoles.get_console_settings_from_registry(
                context=self.context,
                config_path=self.config_path,
                kernel_module_name=self.config["kernel"],
                max_history=max_history,
                max_buffers=[],
            )

        vollog.debug(f"Possible CommandHistorySize values: {max_history}")

        proc = None
        for (
            proc,
            command_history,
            command_history_properties,
        ) in self.get_command_history(
            self.context,
            self.config_path,
            self.config["kernel"],
            procs,
            max_history,
        ):
            process_name = utility.array_to_string(proc.ImageFileName)
            process_pid = proc.UniqueProcessId

            if command_history and command_history_properties:
                for command_history_property in command_history_properties:
                    yield (
                        command_history_property["level"],
                        (
                            process_pid,
                            process_name,
                            format_hints.Hex(command_history.vol.offset),
                            command_history_property["name"],
                            (
                                renderers.NotApplicableValue()
                                if command_history_property["address"] is None
                                else format_hints.Hex(
                                    command_history_property["address"]
                                )
                            ),
                            str(command_history_property["data"]),
                        ),
                    )
            else:
                yield (
                    0,
                    (
                        process_pid,
                        process_name,
                        (
                            format_hints.Hex(command_history.vol.offset)
                            if command_history
                            else renderers.NotApplicableValue()
                        ),
                        "_COMMAND_HISTORY",
                        renderers.NotApplicableValue(),
                        "History Not Found",
                    ),
                )

        if proc is None:
            vollog.warn("No conhost.exe processes found.")

    def _conhost_proc_filter(self, proc: interfaces.objects.ObjectInterface):
        """
        Used to filter only conhost.exe processes
        """
        process_name = utility.array_to_string(proc.ImageFileName)

        return process_name != "conhost.exe"

    def run(self):
        return renderers.TreeGrid(
            [
                ("PID", int),
                ("Process", str),
                ("ConsoleInfo", format_hints.Hex),
                ("Property", str),
                ("Address", format_hints.Hex),
                ("Data", str),
            ],
            self._generator(
                pslist.PsList.list_processes(
                    context=self.context,
                    kernel_module_name=self.config["kernel"],
                    filter_func=self._conhost_proc_filter,
                )
            ),
        )



# This file is Copyright 2020 Volatility Foundation and licensed under the Volatility Software License 1.0
# which is available at https://www.volatilityfoundation.org/license/vsl-v1.0
#
import binascii
import hashlib
import logging
from struct import pack, unpack
from typing import List, Optional, Tuple

try:
    from Crypto.Cipher import ARC4, DES, AES
except ImportError:
    # Debian/Ubuntu ship pycryptodome under Cryptodome namespace
    from Cryptodome.Cipher import ARC4, DES, AES

from volatility3.framework import interfaces, renderers, exceptions, constants
from volatility3.framework.configuration import requirements
from volatility3.framework.layers import registry as registry_layer
from volatility3.framework.symbols.windows.extensions import registry
from volatility3.plugins.windows.registry import hivelist

vollog = logging.getLogger(__name__)


class Hashdump(interfaces.plugins.PluginInterface):
    """Dumps user hashes from memory"""

    _required_framework_version = (2, 0, 0)
    _version = (1, 1, 1)

    @classmethod
    def get_requirements(cls):
        return [
            requirements.ModuleRequirement(
                name="kernel",
                description="Windows kernel",
                architectures=["Intel32", "Intel64"],
            ),
            requirements.VersionRequirement(
                name="hivelist", component=hivelist.HiveList, version=(2, 0, 0)
            ),
        ]

    odd_parity = [
        1,
        1,
        2,
        2,
        4,
        4,
        7,
        7,
        8,
        8,
        11,
        11,
        13,
        13,
        14,
        14,
        16,
        16,
        19,
        19,
        21,
        21,
        22,
        22,
        25,
        25,
        26,
        26,
        28,
        28,
        31,
        31,
        32,
        32,
        35,
        35,
        37,
        37,
        38,
        38,
        41,
        41,
        42,
        42,
        44,
        44,
        47,
        47,
        49,
        49,
        50,
        50,
        52,
        52,
        55,
        55,
        56,
        56,
        59,
        59,
        61,
        61,
        62,
        62,
        64,
        64,
        67,
        67,
        69,
        69,
        70,
        70,
        73,
        73,
        74,
        74,
        76,
        76,
        79,
        79,
        81,
        81,
        82,
        82,
        84,
        84,
        87,
        87,
        88,
        88,
        91,
        91,
        93,
        93,
        94,
        94,
        97,
        97,
        98,
        98,
        100,
        100,
        103,
        103,
        104,
        104,
        107,
        107,
        109,
        109,
        110,
        110,
        112,
        112,
        115,
        115,
        117,
        117,
        118,
        118,
        121,
        121,
        122,
        122,
        124,
        124,
        127,
        127,
        128,
        128,
        131,
        131,
        133,
        133,
        134,
        134,
        137,
        137,
        138,
        138,
        140,
        140,
        143,
        143,
        145,
        145,
        146,
        146,
        148,
        148,
        151,
        151,
        152,
        152,
        155,
        155,
        157,
        157,
        158,
        158,
        161,
        161,
        162,
        162,
        164,
        164,
        167,
        167,
        168,
        168,
        171,
        171,
        173,
        173,
        174,
        174,
        176,
        176,
        179,
        179,
        181,
        181,
        182,
        182,
        185,
        185,
        186,
        186,
        188,
        188,
        191,
        191,
        193,
        193,
        194,
        194,
        196,
        196,
        199,
        199,
        200,
        200,
        203,
        203,
        205,
        205,
        206,
        206,
        208,
        208,
        211,
        211,
        213,
        213,
        214,
        214,
        217,
        217,
        218,
        218,
        220,
        220,
        223,
        223,
        224,
        224,
        227,
        227,
        229,
        229,
        230,
        230,
        233,
        233,
        234,
        234,
        236,
        236,
        239,
        239,
        241,
        241,
        242,
        242,
        244,
        244,
        247,
        247,
        248,
        248,
        251,
        251,
        253,
        253,
        254,
        254,
    ]

    # Permutation matrix for boot key
    bootkey_perm_table = [
        0x8,
        0x5,
        0x4,
        0x2,
        0xB,
        0x9,
        0xD,
        0x3,
        0x0,
        0x6,
        0x1,
        0xC,
        0xE,
        0xA,
        0xF,
        0x7,
    ]

    # Constants for SAM decrypt algorithm
    aqwerty = b"!@#$%^&*()qwertyUIOPAzxcvbnmQQQQQQQQQQQQ)(*@&%\0"
    anum = b"0123456789012345678901234567890123456789\0"
    antpassword = b"NTPASSWORD\0"
    almpassword = b"LMPASSWORD\0"
    lmkey = b"KGS!@#$%"

    empty_lm = b"\xaa\xd3\xb4\x35\xb5\x14\x04\xee\xaa\xd3\xb4\x35\xb5\x14\x04\xee"
    empty_nt = b"\x31\xd6\xcf\xe0\xd1\x6a\xe9\x31\xb7\x3c\x59\xd7\xe0\xc0\x89\xc0"

    @classmethod
    def get_hive_key(
        cls, hive: registry_layer.RegistryHive, key: str
    ) -> Optional["registry.CM_KEY_NODE"]:
        result = None
        try:
            if hive:
                result = hive.get_key(key)
        except (KeyError, registry_layer.RegistryException):
            vollog.info(
                f"Unable to load the required registry key {hive.get_name()}\\{key} from this memory image"
            )
        return result

    @classmethod
    def get_user_keys(
        cls, samhive: registry_layer.RegistryHive
    ) -> List[interfaces.objects.ObjectInterface]:
        user_key_path = "SAM\\Domains\\Account\\Users"

        user_key = cls.get_hive_key(samhive, user_key_path)

        if not user_key:
            return []
        return [k for k in user_key.get_subkeys() if k.get_name() != "Names"]

    @classmethod
    def get_bootkey(cls, syshive: registry_layer.RegistryHive) -> Optional[bytes]:
        """
        Returns the scrambled bootkey necessary to decrypt hashes
        """
        cs = 1
        lsa_base = f"ControlSet{cs:03}" + "\\Control\\Lsa"
        lsa_keys = ["JD", "Skew1", "GBG", "Data"]

        lsa = cls.get_hive_key(syshive, lsa_base)
        if not lsa:
            return None

        bootkey = ""

        for lk in lsa_keys:
            try:
                key = cls.get_hive_key(syshive, lsa_base + "\\" + lk)
                class_data = None
                if key:
                    try:
                        class_data = syshive.read(key.Class + 4, key.ClassLength)
                    except exceptions.InvalidAddressException:
                        return None

                if class_data is None:
                    return None
                bootkey += class_data.decode("utf-16-le")
            except (
                exceptions.InvalidAddressException,
                registry_layer.RegistryException,
            ) as excp:
                vollog.log(
                    constants.LOGLEVEL_VVV, f"Unable to read Lsa key {lk}: {excp}"
                )
                return None

        bootkey_str = binascii.unhexlify(bootkey)
        bootkey_scrambled = bytes(
            [bootkey_str[cls.bootkey_perm_table[i]] for i in range(len(bootkey_str))]
        )
        return bootkey_scrambled

    @classmethod
    def get_hbootkey(
        cls, samhive: registry_layer.RegistryHive, bootkey: bytes
    ) -> Optional[bytes]:
        sam_account_path = "SAM\\Domains\\Account"

        if not bootkey:
            return None

        sam_account_key = cls.get_hive_key(samhive, sam_account_path)
        if not sam_account_key:
            return None

        sam_data = None
        for v in sam_account_key.get_values():
            if v.get_name() == "F":
                try:
                    sam_data = samhive.read(v.Data + 4, v.DataLength)
                except exceptions.InvalidAddressException:
                    return None

        if not sam_data:
            return None

        revision = sam_data[0x00]
        if revision == 2:
            md5 = hashlib.md5()

            md5.update(sam_data[0x70:0x80] + cls.aqwerty + bootkey + cls.anum)
            rc4_key = md5.digest()

            rc4 = ARC4.new(rc4_key)
            hbootkey = rc4.encrypt(
                sam_data[0x80:0xA0]
            )  # lgtm [py/weak-cryptographic-algorithm]
            return hbootkey
        elif revision == 3:
            # AES encrypted
            iv = sam_data[0x78:0x88]
            encryptedHBootKey = sam_data[0x88:0xA8]
            cipher = AES.new(bootkey, AES.MODE_CBC, iv)
            hbootkey = cipher.decrypt(encryptedHBootKey)
            return hbootkey[:16]
        return None

    @classmethod
    def decrypt_single_salted_hash(
        cls, rid, hbootkey: bytes, enc_hash: bytes, _lmntstr, salt: bytes
    ) -> Optional[bytes]:
        (des_k1, des_k2) = cls.sid_to_key(rid)
        des1 = DES.new(des_k1, DES.MODE_ECB)
        des2 = DES.new(des_k2, DES.MODE_ECB)
        cipher = AES.new(hbootkey[:16], AES.MODE_CBC, salt)
        obfkey = cipher.decrypt(enc_hash)
        return des1.decrypt(obfkey[:8]) + des2.decrypt(
            obfkey[8:16]
        )  # lgtm [py/weak-cryptographic-algorithm]

    @classmethod
    def get_user_hashes(
        cls,
        user: registry.CM_KEY_NODE,
        samhive: registry_layer.RegistryHive,
        hbootkey: bytes,
    ) -> Optional[Tuple[bytes, bytes]]:
        ## Will sometimes find extra user with rid = NAMES, returns empty strings right now
        try:
            rid = int(str(user.get_name()), 16)
        except ValueError:
            return None
        sam_data = None
        for v in user.get_values():
            if v.get_name() == "V":
                try:
                    sam_data = samhive.read(v.Data + 4, v.DataLength)
                except (
                    exceptions.InvalidAddressException,
                    registry_layer.RegistryException,
                ):
                    return None

        if not sam_data:
            return None

        lm_offset = unpack("<L", sam_data[0x9C:0xA0])[0] + 0xCC
        lm_len = unpack("<L", sam_data[0xA0:0xA4])[0]
        nt_offset = unpack("<L", sam_data[0xA8:0xAC])[0] + 0xCC
        nt_len = unpack("<L", sam_data[0xAC:0xB0])[0]

        lm_revision = sam_data[lm_offset + 2 : lm_offset + 3]
        lmhash = None
        if lm_revision == b"\x01":
            if lm_len == 20:
                enc_lm_hash = sam_data[lm_offset + 0x04 : lm_offset + 0x14]
                lmhash = cls.decrypt_single_hash(
                    rid, hbootkey, enc_lm_hash, cls.almpassword
                )
        elif lm_revision == b"\x02":
            if lm_len == 56:
                lm_salt = sam_data[lm_offset + 4 : lm_offset + 20]
                enc_lm_hash = sam_data[lm_offset + 20 : lm_offset + 52]
                lmhash = cls.decrypt_single_salted_hash(
                    rid, hbootkey, enc_lm_hash, cls.almpassword, lm_salt
                )

        # NT hash decryption
        nthash = None
        nt_revision = sam_data[nt_offset + 2 : nt_offset + 3]
        if nt_revision == b"\x01":
            if nt_len == 20:
                enc_nt_hash = sam_data[nt_offset + 4 : nt_offset + 20]
                nthash = cls.decrypt_single_hash(
                    rid, hbootkey, enc_nt_hash, cls.antpassword
                )
        elif nt_revision == b"\x02":
            if nt_len == 56:
                nt_salt = sam_data[nt_offset + 8 : nt_offset + 24]
                enc_nt_hash = sam_data[nt_offset + 24 : nt_offset + 56]
                nthash = cls.decrypt_single_salted_hash(
                    rid, hbootkey, enc_nt_hash, cls.antpassword, nt_salt
                )
        return lmhash, nthash

    @classmethod
    def sid_to_key(cls, sid: int) -> Tuple[bytes, bytes]:
        """Takes rid of a user and converts it to a key to be used by the DES cipher"""
        bytestr1 = [
            sid & 0xFF,
            (sid >> 8) & 0xFF,
            (sid >> 16) & 0xFF,
            (sid >> 24) & 0xFF,
        ]
        bytestr1 += bytestr1[0:3]
        bytestr2 = [bytestr1[3]] + bytestr1[0:3]
        bytestr2 += bytestr2[0:3]
        return cls.sidbytes_to_key(bytes(bytestr1)), cls.sidbytes_to_key(
            bytes(bytestr2)
        )

    @classmethod
    def sidbytes_to_key(cls, s: bytes) -> bytes:
        """Builds final DES key from the strings generated in sid_to_key"""
        key = [
            s[0] >> 1,
            ((s[0] & 0x01) << 6) | (s[1] >> 2),
            ((s[1] & 0x03) << 5) | (s[2] >> 3),
            ((s[2] & 0x07) << 4) | (s[3] >> 4),
            ((s[3] & 0x0F) << 3) | (s[4] >> 5),
            ((s[4] & 0x1F) << 2) | (s[5] >> 6),
            ((s[5] & 0x3F) << 1) | (s[6] >> 7),
            s[6] & 0x7F,
        ]
        for i in range(8):
            key[i] = key[i] << 1
            key[i] = cls.odd_parity[key[i]]
        return bytes(key)

    @classmethod
    def decrypt_single_hash(
        cls, rid: int, hbootkey: bytes, enc_hash: bytes, lmntstr: bytes
    ):
        (des_k1, des_k2) = cls.sid_to_key(rid)
        des1 = DES.new(des_k1, DES.MODE_ECB)
        des2 = DES.new(des_k2, DES.MODE_ECB)
        md5 = hashlib.md5()

        md5.update(hbootkey[:0x10] + pack("<L", rid) + lmntstr)
        rc4_key = md5.digest()
        rc4 = ARC4.new(rc4_key)
        obfkey = rc4.encrypt(enc_hash)  # lgtm [py/weak-cryptographic-algorithm]

        return des1.decrypt(obfkey[:8]) + des2.decrypt(
            obfkey[8:]
        )  # lgtm [py/weak-cryptographic-algorithm]

    @classmethod
    def get_user_name(
        cls, user: registry.CM_KEY_NODE, samhive: registry_layer.RegistryHive
    ) -> Optional[bytes]:
        value = None
        for v in user.get_values():
            if v.get_name() == "V":
                try:
                    value = samhive.read(v.Data + 4, v.DataLength)
                except exceptions.InvalidAddressException:
                    return None

        if not value:
            return None

        name_offset = unpack("<L", value[0x0C:0x10])[0] + 0xCC
        name_length = unpack("<L", value[0x10:0x14])[0]
        if name_length > len(value):
            return None

        username = value[name_offset : name_offset + name_length]
        return username

    # replaces the dump_hashes method in vol2
    def _generator(
        self, syshive: registry_layer.RegistryHive, samhive: registry_layer.RegistryHive
    ):
        if syshive is None:
            vollog.debug("SYSTEM address is None: No system hive found")
        if samhive is None:
            vollog.debug("SAM address is None: No SAM hive found")
        bootkey = self.get_bootkey(syshive)
        hbootkey = self.get_hbootkey(samhive, bootkey)
        if hbootkey:
            for user in self.get_user_keys(samhive):
                ret = self.get_user_hashes(user, samhive, hbootkey)
                if ret:
                    lmhash, nthash = ret

                    ## temporary fix to prevent UnicodeDecodeError backtraces
                    ## however this can cause truncated user names as a result
                    name = self.get_user_name(user, samhive)
                    if name is None:
                        name = renderers.NotAvailableValue()
                    else:
                        name = str(name, "utf-16-le", errors="ignore")

                    lmout = str(binascii.hexlify(lmhash or self.empty_lm), "latin-1")
                    ntout = str(binascii.hexlify(nthash or self.empty_nt), "latin-1")
                    rid = int(str(user.get_name()), 16)
                    yield (0, (name, rid, lmout, ntout))
        else:
            vollog.warning("Hbootkey is not valid")

    def run(self):
        offset = self.config.get("offset", None)
        syshive = None
        samhive = None
        for hive in hivelist.HiveList.list_hives(
            context=self.context,
            base_config_path=self.config_path,
            kernel_module_name=self.config["kernel"],
            hive_offsets=None if offset is None else [offset],
        ):
            if hive.get_name().split("\\")[-1].upper() == "SYSTEM":
                syshive = hive
            if hive.get_name().split("\\")[-1].upper() == "SAM":
                samhive = hive

        return renderers.TreeGrid(
            [("User", str), ("rid", int), ("lmhash", str), ("nthash", str)],
            self._generator(syshive, samhive),
        )
